# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from . import files as base
from .__about__ import __version__
from .__regex import RegexConf
from .config import (
    UPDATE_KEY,
    VERSION_KEY,
    Params,
    Paths,
    Rule,
    Stage,
)
from .exceptions import (
    IOBaseError,
    StoreArgumentError,
    StoreNotFound,
)
from .files import (
    CsvFl,
    CsvPipeFl,
    EnvFl,
    Fl,
    JsonEnvFl,
    JsonFl,
    JsonLineFl,
    MarshalFl,
    MsgpackFl,
    PickleFl,
    TomlEnvFl,
    TomlFl,
    YamlEnvFl,
    YamlFl,
    YamlFlResolve,
)
from .paths import (
    PathSearch,
    glob_files,
    is_ignored,
    ls,
    read_ignore,
    replace_sep,
)
from .stores import (
    BaseStore,
    Store,
    StoreJsonToCsv,
    StoreToJsonLine,
)
from .utils import (
    map_func,
    rm,
    search_env,
    search_env_replace,
    template_func,
    template_secret,
    touch,
)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
"""
This is the main function for open any files in local or remote space
with the best python libraries and the best practice such as build-in
``io.open``, ``mmap.mmap``, etc.

TODO: Add more compress type such as
    - h5,hdf5(h5py)
    - fits(astropy)
    - rar(...)
"""
from __future__ import annotations

import abc
import csv
import io
import json
import logging
import marshal
import mmap
import os
import pickle
import re
from contextlib import contextmanager
from pathlib import Path
from typing import (
    IO,
    Any,
    AnyStr,
    Callable,
    ClassVar,
    Literal,
    Optional,
    Protocol,
    Union,
    get_args,
)

try:
    import msgpack
except ImportError:  # pragma: no cov
    msgpack = None

try:
    import yaml

    try:
        from yaml import CSafeLoader as SafeLoader
        from yaml import CUnsafeLoader as UnsafeLoader
    except ImportError:  # pragma: no cov
        from yaml import SafeLoader, UnsafeLoader
except ImportError:  # pragma: no cov
    yaml = None
    SafeLoader = None
    UnsafeLoader = None

try:
    import toml
except ImportError:  # pragma: no cov
    toml = None

try:
    import tomllib
except ModuleNotFoundError:  # pragma: no cov
    import pip._vendor.tomli as tomllib

from .utils import search_env, search_env_replace

logger = logging.getLogger("ddeutil.io")
FileCompressType = Literal["gzip", "gz", "xz", "bz2"]

__all__: tuple[str, ...] = (
    "Fl",
    "EnvFlMixin",
    "EnvFl",
    "JsonFl",
    "JsonEnvFl",
    "JsonLineFl",
    "YamlFl",
    "YamlFlResolve",
    "YamlEnvFl",
    "CsvFl",
    "CsvPipeFl",
    "TomlFl",
    "TomlEnvFl",
    "MarshalFl",
    "MsgpackFl",
    "PickleFl",
    "compress_lib",
)


class CompressProtocol(Protocol):  # pragma: no cov
    """Compress protocol object that allow to implement and use ``decompress``
    and ``open`` methods.
    """

    def decompress(self, *args, **kwargs) -> AnyStr: ...

    def open(self, *args, **kwargs) -> IO: ...


def compress_lib(compress: Optional[FileCompressType]) -> CompressProtocol:
    """Return Compress module that use to unpack data from the compressed file.
    Now, it support for "gzip", "gz", "xz", and "bz2".

    :param compress: A compress string type value that want to get compress
        package.
    :type compress: str
    :rtype: CompressProtocol
    """
    if not compress:
        return io
    elif compress in ("gzip", "gz"):
        import gzip

        return gzip
    elif compress in ("bz2",):
        import bz2

        return bz2
    elif compress in ("xz",):
        import lzma as xz

        return xz
    raise NotImplementedError(f"Compress {compress} does not implement yet")


class FlABC(abc.ABC):  # pragma: no cov
    """Open File abstraction object for marking abstract methods that need to
    implement on any open file subclass.
    """

    @abc.abstractmethod
    def read(self, *args, **kwargs): ...

    @abc.abstractmethod
    def write(self, *args, **kwargs): ...


class Fl(FlABC):
    """Open File object that use to open any normal or compression file from
    current local file system (I do not have plan to implement remote object
    storage like AWS S3, GCS, or ADLS).

        Note that, this object should to implement it with subclass again
    because it does not override necessary methods from FlABC abstract class.

    :param path: A path that respresent the file location.
    :type path: str | Path
    :param encoding: An open file encoding value, it will use UTF-8 by default.
    :type encoding: Optional[str] (None)
    :param compress: A compress type for this file.
    :type compress: FileCompressType | None (None)

    Examples:
        >>> with Fl(
        ...     path='./<path>/<filename>.gz.txt', compress='gzip'
        ... ).open() as f:
        ...     data = f.readline()
    """

    def __init__(
        self,
        path: Union[str, Path],
        *,
        encoding: Optional[str] = None,
        compress: Optional[FileCompressType] = None,
    ) -> None:
        self.path: Path = Path(path) if isinstance(path, str) else path
        self.encoding: str = encoding or "utf-8"
        self.compress: Optional[FileCompressType] = compress

        # NOTE: Action anything after set up attributes.
        self.after_set_attrs()

    def after_set_attrs(self) -> None:  # pragma: no cov
        """Do any action after the object initialize step."""

    def __call__(self, *args, **kwargs) -> IO:
        """Return IO of this object."""
        return self.open(*args, **kwargs)

    @property
    def decompress(self) -> Callable[[...], AnyStr]:
        """Return decompress method that getting from its compression type.

        :rtype: Callable[[...], AnyStr]
        """
        if self.compress is not None and self.compress in get_args(
            FileCompressType
        ):
            return compress_lib(self.compress).decompress
        raise NotImplementedError(
            "Does not implement decompress method for None compress value."
        )

    def __mode(self, mode: Optional[str] = None) -> dict[str, str]:
        """Convert mode property before passing to the main standard lib.

        :param mode: a reading or writing mode for the open method.
        :type mode: Optional[str] (None)

        :rtype: dict[str, str]
        :returns: A mapping of mode and other input parameters for standard
            libs.
        """
        if not mode:
            return {"mode": "r"}

        byte_mode: bool = "b" in mode
        if self.compress is None:
            _mode: dict[str, str] = {"mode": mode}
            return _mode if byte_mode else {"encoding": self.encoding, **_mode}

        if self.compress in get_args(FileCompressType):
            return (
                {"mode": mode}
                if byte_mode
                else {"mode": f"{mode}t", "encoding": self.encoding}
            )

        raise NotImplementedError(
            f"mode conversion does not support for compress type no in "
            f"{get_args(FileCompressType)}."
        )

    def open(self, *, mode: Optional[str] = None, **kwargs) -> IO:
        """Open this file object with standard libs that match with it file
        format subclass propose.

        :param mode: An opening mode that allow you to use read or write mode.
        :type mode: Optional[str] (None)
        :rtype: IO
        """
        return compress_lib(self.compress).open(
            self.path, **(self.__mode(mode) | kwargs)
        )

    @contextmanager
    def mopen(self, *, mode: Optional[str] = None) -> IO:
        """Open with memory mode context manager.

        :param mode: A opening mode that allow you to use read or write mode.
        :type mode: Optional[str] (None)
        :rtype: IO
        """
        mode: str = mode or "r"
        file: IO = self.open(mode=mode)
        _access: int = mmap.ACCESS_READ if ("r" in mode) else mmap.ACCESS_WRITE
        try:
            yield mmap.mmap(file.fileno(), length=0, access=_access)
        except ValueError:
            logger.exception("Can not open file with memory mode")
            yield file
        finally:
            file.close()

    def read(self, *args, **kwargs):  # pragma: no cov
        raise NotImplementedError(
            "This is abstract class only, so, you should implement open file "
            "object with this class and override this method."
        )

    def write(self, *args, **kwargs) -> None:  # pragma: no cov
        raise NotImplementedError(
            "This is abstract class only, so, you should implement open file "
            "object with this class and override this method."
        )


class EnvFlMixin:
    """Environment Mapping to read method of open file object mixin. This object
    already implement class variables that need to use on ``search_env_replace``
    function.
    """

    raise_if_not_default: ClassVar[bool] = False
    default: ClassVar[str] = "null"
    escape: ClassVar[str] = "<ESCAPE>"

    @staticmethod
    def prepare(value: str) -> str:
        """Prepare function it use on searching environment variable process
        that passing string value to this function before keeping to the final
        context data.

        :param value: A string value that passing from searching process
        :type value: str

        :rtype: str
        """
        return value

    def search_env_replace(self, content: str) -> str:
        """Return environment variable replaced content.

        :param content: A content data that want to search and replace env var.
        """
        return search_env_replace(
            content,
            raise_if_default_not_exists=self.raise_if_not_default,
            default=self.default,
            escape=self.escape,
            caller=self.prepare,
        )


class EnvFl(Fl):
    """Dot env open file object which mapping search engine to data context that
    reading from dot env file format (.env).
    """

    keep_newline: ClassVar[bool] = False
    default: ClassVar[str] = ""

    def read(self, *, update: bool = True) -> dict[str, str]:
        """Return data context from dot env file format.

        :param update: A update environment variable to interpreter flag.
        :type update: bool (True)
        :rtype: dict[str, str]
        """
        with self.open(mode="r") as f:
            f.seek(0)
            rs: dict[str, str] = search_env(
                f.read(),
                keep_newline=self.keep_newline,
                default=self.default,
            )
        if update:
            os.environ.update(**rs)
        return rs

    def write(self, data: dict[str, Any]) -> None:  # pragma: no cov
        raise NotImplementedError(
            "Dot env open file object does not allow to write."
        )


class YamlFl(Fl):
    """Yaml open file object that read data context from Yaml file format (.yml,
    or .yaml).

        Note that, the boolean values on the data context in the yaml file will
    convert to the Python object;
        * true:     y, Y, true, Yes, on, ON
        * false:    n, N, false, No, off, OFF
    """

    def read(self, safe: bool = True) -> dict[str, Any]:
        """Return data context from yaml file format.

        :param safe: A flag that allow to use safe reading mode.
        :type safe: bool (True)
        :rtype: dict[str, Any]
        """
        with self.open(mode="r") as f:
            return yaml.load(f.read(), (SafeLoader if safe else UnsafeLoader))

    def write(self, data: dict[str, Any]) -> None:
        with self.open(mode="w") as f:
            yaml.dump(data, f, default_flow_style=False)


class YamlFlResolve(YamlFl):
    """Yaml open file object with resolve boolean convert value problem such as
    convert 'on' value to true instead a string of 'on' value. This object also
    read data context from Yaml file format (.yml, or .yaml).
    """

    def read(self, safe: bool = True) -> dict[str, Any]:
        """Reading Yaml data with does not convert boolean value.

        :param safe: A flag that allow to use safe reading mode.
        :type safe: bool (True)
        :rtype: dict[str, Any]

        Note that:
            Handle top level yaml property ``on``
            docs: https://github.com/yaml/pyyaml/issues/696

            ```
            import re
            from yaml.resolver import Resolver

            # zap the Resolver class' internal dispatch table
            Resolver.yaml_implicit_resolvers = {}

            # NOTE: Current Resolver
            Resolver.add_implicit_resolver(
                    'tag:yaml.org,2002:bool',
                    re.compile(r'''^(?:yes|Yes|YES|no|No|NO
                                |true|True|TRUE|false|False|FALSE
                                |on|On|ON|off|Off|OFF)$''', re.X),
                    list('yYnNtTfFoO'))

            # NOTE: The 1.2 bool impl Resolver:
            Resolver.add_implicit_resolver(
                    'tag:yaml.org,2002:bool',
                    re.compile(r'^(?:true|false)$', re.X),
                    list('tf'))
            ```
        """
        from yaml.resolver import Resolver

        revert = Resolver.yaml_implicit_resolvers.copy()

        for ch in "OoYyNn":
            if len(Resolver.yaml_implicit_resolvers[ch]) == 1:
                del Resolver.yaml_implicit_resolvers[ch]
            else:
                Resolver.yaml_implicit_resolvers[ch] = [
                    x
                    for x in Resolver.yaml_implicit_resolvers[ch]
                    if x[0] != "tag:yaml.org,2002:bool"
                ]

        with self.open(mode="r") as f:
            rs: dict[str, Any] = yaml.load(
                f.read(), (SafeLoader if safe else UnsafeLoader)
            )
            # NOTE: Override revert resolver when want to use safe load.
            Resolver.yaml_implicit_resolvers = revert
            return rs


class YamlEnvFl(YamlFl, EnvFlMixin):
    """Yaml open file object which mapping search environment variable."""

    def read(self, safe: bool = True) -> dict[str, Any]:
        """Return data context from yaml file format and mapping search
        environment variables before returning context data.

        :param safe: A flag that allow to use safe reading mode.
        :type safe: bool (True)
        :rtype: dict[str, Any]
        """
        with self.open(mode="r") as f:
            context_env_replace: str = self.search_env_replace(
                yaml.dump(yaml.load(f.read(), UnsafeLoader))
            )
        return yaml.load(
            context_env_replace,
            (SafeLoader if safe else UnsafeLoader),
        )

    def write(self, data: dict[str, Any]) -> None:  # pragma: no cov
        raise NotImplementedError(
            "Yaml open file with mapping env var does not allow to write."
        )


class CsvFl(Fl):
    """CSV open file object with comma (,) seperator charactor."""

    def read(self, pre_load: int = 0) -> list[dict[Union[str, int], Any]]:
        """Return data context from csv file format.

        :param pre_load: An input bytes number that use to preloading for
            define column structure before reading with csv.
        :type pre_load: int (0)
        :rtype: list[dict[str | int, Any]]
        """
        with self.open(mode="r") as f:
            return list(csv.DictReader(f, delimiter=",", quoting=csv.QUOTE_ALL))

    def write(
        self,
        data: Union[list[Any], dict[Any, Any]],
        *,
        mode: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Write CSV file with an input data context. This method allow to use
        append write mode.
        """
        if not data:
            raise ValueError("data that want writing to CSV file was empty")

        mode: str = mode or "w"
        assert mode in (
            "a",
            "w",
        ), "save mode in CSV must contain only value `a` nor `w`."

        if isinstance(data, dict):
            data: list = [data]

        with self.open(mode=mode, newline="") as f:
            # noinspection PyTypeChecker
            writer = csv.DictWriter(
                f,
                fieldnames=list(data[0].keys()),
                lineterminator="\n",
                **kwargs,
            )
            if mode == "w" or not self.has_header:
                writer.writeheader()
            writer.writerows(data)

    @property
    def has_header(self, pre_load: int = 128) -> bool:
        """Return true if the file with csv format already implement header.

        :param pre_load: An input bytes number that use to preloading for
            define header.
        :type pre_load: int (128)
        :rtype: bool
        """
        with self.open(mode="r") as f:
            try:
                return csv.Sniffer().has_header(f.read(pre_load))
            except csv.Error:
                return False


class CsvDynamicFl(CsvFl):  # pragma: no cov
    """CSV open file object with dynamic dialect reader."""

    def read(self, pre_load: int = 128) -> list[dict[Union[str, int], Any]]:
        """Return data context from csv file format with dynamic dialect reader.

        :param pre_load: An input bytes number that use to preloading for
            define column structure before reading with csv.
        :type pre_load: int (128)
        :rtype: list[dict[str | int, Any]]
        """
        with self.open(mode="r") as f:
            dialect = csv.Sniffer().sniff(f.read(pre_load))
            f.seek(0)
            return list(csv.DictReader(f, dialect=dialect))


class CsvPipeFl(CsvFl):
    """CSV open file object with pipe (|) seperator charactor."""

    def after_set_attrs(self) -> None:
        """Register csv dialect after setting attribute open file object."""
        csv.register_dialect(
            "pipe_delimiter", delimiter="|", quoting=csv.QUOTE_ALL
        )

    def read(self, pre_load: int = 0) -> list:
        """Return data context from csv file format with the pipe seperator.
        This read method do not use ``pre_load`` parameter because it is passing
        fix dialect argument.

        :param pre_load: An input bytes number that use to preloading for
            define column structure before reading with csv.
        :type pre_load: int (0)
        :rtype: list[dict[str | int, Any]]
        """
        with self.open(mode="r") as f:
            return list(csv.DictReader(f, delimiter="|", quoting=csv.QUOTE_ALL))

    def write(
        self,
        data: Union[list[Any], dict[Any, Any]],
        *,
        mode: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Write data to the CSV file format."""
        if not data:
            raise ValueError("data to write is empty")

        mode: str = mode or "w"
        assert mode in {
            "a",
            "w",
        }, "save mode must contain only value `a` nor `w`."

        if isinstance(data, dict):
            data: list = [data]

        with self.open(mode=mode, newline="") as f:
            # noinspection PyTypeChecker
            writer = csv.DictWriter(
                f,
                fieldnames=list(data[0].keys()),
                lineterminator="\n",
                delimiter="|",
                quoting=csv.QUOTE_ALL,
                **kwargs,
            )
            if mode == "w" or not self.has_header:
                writer.writeheader()
            writer.writerows(data)


class JSONCommentsDecoder(json.JSONDecoder):
    """Override JSON Decoder object for remove comment statement inside data
    content that not remove by default json built-in module.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def decode(self, s: str, _w=None):
        """Decode with a comment in the Json data context.

        Reference:
            - https://stackoverflow.com/questions/69021815/
                how-to-read-json-file-with-comments
        """
        # NOTE: This below regex requires the ``re.X`` flag, to ignore
        #   whitespaces inside the regex.
        #
        #   (
        #       "(?:\\"|[^"])*?" (?# First try to match balanced quotes)
        #   )
        #   | (?# If not successfully, try the following)
        #   (
        #       \/\*(?:.|\s)*?\*\/ (?# Match a block comment)
        #       |
        #       \/\/.* (?# Match a line comment)
        #   )
        #
        regex: str = r"""("(?:\\"|[^"])*?")|(\/\*(?:.|\s)*?\*\/|\/\/.*)"""
        s: str = re.sub(regex, r"\1", s)  # NOTE: , flags = re.X | re.M)
        return super().decode(s)


class JsonFl(Fl):
    """Json open file object that read data context from Json file format
    (.json).
    """

    def read(self) -> Union[dict[Any, Any], list[Any]]:
        with self.open(mode="r") as f:
            try:
                return json.loads(f.read(), cls=JSONCommentsDecoder)
            except json.decoder.JSONDecodeError as err:
                logger.exception(err)
                raise

    def write(self, data, *, indent: int = 4) -> None:
        with self.open(mode="w") as f:
            if self.compress:
                f.write(json.dumps(data, default=str))
            else:
                # noinspection PyTypeChecker
                json.dump(data, f, indent=indent, default=str)


class JsonEnvFl(JsonFl, EnvFlMixin):
    """Json open file object which mapping search environment variable before
    parsing with json package.
    """

    def read(self) -> Union[dict[Any, Any], list[Any]]:
        with self.open(mode="rt") as f:
            try:
                return json.loads(
                    self.search_env_replace(f.read()),
                    cls=JSONCommentsDecoder,
                )
            except json.decoder.JSONDecodeError as err:
                logger.exception(err)
                raise

    def write(self, data, *, indent: int = 4) -> None:  # pragma: no cov
        raise NotImplementedError(
            "Json open file with mapping env var does not allow to write."
        )


class JsonLineFl(Fl):
    """Json open file object that read data context from Json file format
    (.json) with a newline seperator.
    """

    def read(self) -> list[Any]:
        rs: list[Any] = []
        with self.open(mode="rt") as f:
            for line in f:
                try:
                    rs.append(json.loads(line, cls=JSONCommentsDecoder))
                except json.decoder.JSONDecodeError as err:
                    logger.exception(err)
                    raise
        return rs

    def write(self, data, *, mode: Optional[str] = None) -> None:
        if not data:
            raise ValueError("data to write is empty")

        mode: str = mode or "w"
        assert mode in {
            "a",
            "w",
        }, "save mode must contain only value `a` nor `w`."

        with self.open(mode=mode) as f:
            if isinstance(data, list):
                for d in data:
                    f.write(json.dumps(d, default=str) + "\n")
            else:
                f.write(json.dumps(data, default=str) + "\n")


class TomlFl(Fl):
    """TOML open file object that read data context from TOML file format
    (.toml).
    """

    def read(self):
        with self.open(mode="rt") as f:
            return tomllib.loads(f.read())

    def write(self, data: dict[str, Any]) -> None:
        if toml is None:  # pragma: no cov
            raise ImportError(
                "writing toml file need `toml` package, you should to install "
                "toml via `pip install toml` first."
            )
        with self.open(mode="wt") as f:
            # noinspection PyTypeChecker
            toml.dump(data, f)


class TomlEnvFl(TomlFl, EnvFlMixin):
    """TOML open file object which mapping search environment variable before
    parsing with toml package from TOML file format (.toml).
    """

    def read(self):
        with self.open(mode="rt") as f:
            return tomllib.loads(self.search_env_replace(f.read()))

    def write(self, data: dict[str, Any]) -> None:  # pragma: no cov
        raise NotImplementedError(
            "Toml open file with mapping env var does not allow to write."
        )


class PickleFl(Fl):  # pragma: no cov
    """Pickle open file object that read data context from Pickle file format
    (.pickle).
    """

    def read(self):
        with self.open(mode="rb") as f:
            return pickle.loads(f.read())

    def write(self, data):
        with self.open(mode="wb") as f:
            # noinspection PyTypeChecker
            pickle.dump(data, f)


class MarshalFl(Fl):  # pragma: no cov
    """Marshal open file object that read data context from Marshal file format.

    Note: use marshal package
    """

    def read(self):
        with self.open(mode="rb") as f:
            return marshal.loads(f.read())

    def write(self, data):
        with self.open(mode="wb") as f:
            # noinspection PyTypeChecker
            marshal.dump(data, f)


class MsgpackFl(Fl):  # pragma: no cov
    """Msgpack open file object that read data context from Msgpack file format.

    Note: use msgpack package
    """

    def read(self):
        with self.open(mode="rb") as f:
            return msgpack.loads(f.read())

    def write(self, data):
        with self.open(mode="wb") as f:
            msgpack.dump(data, f)



import shutil
from collections.abc import Iterator
from pathlib import Path

import pytest
from ddeutil.io.files import Fl
from ddeutil.io.utils import add_newline


@pytest.fixture(scope="module")
def target_path(test_path) -> Iterator[Path]:
    this_path: Path = test_path / "files"
    this_path.mkdir(parents=True, exist_ok=True)

    yield this_path

    shutil.rmtree(this_path)


def test_open_file_common(target_path, encoding):
    file = Fl(
        path=target_path / "test_common_file.text",
        encoding=encoding,
    )
    with file.open(mode="w") as f:
        f.write("Write data with common file in normal mode")

    with file.open(mode="r") as f:
        rs = f.read()

    assert "Write data with common file in normal mode" == rs


def test_open_file_common_with_open(target_path, encoding):
    file = Fl(
        path=target_path / "test_common_file.text",
        encoding=encoding,
    )
    with file.open(mode="w") as f:
        f.write("Write data with common file in normal mode")

    with file() as f:
        rs = f.read()

    assert "Write data with common file in normal mode" == rs


def test_open_file_common_append(target_path, encoding):
    file = Fl(
        path=target_path / "test_common_file_append.text",
        encoding=encoding,
    )
    with file.open(mode="w") as f:
        f.write(
            add_newline(
                "Write data with common file append in normal mode",
            )
        )

    with file.open(mode="a", newline="\n") as f:
        f.write("Write another line in the same file")

    with file.open(mode="r") as f:
        rs = f.read()

    assert (
        "Write data with common file append in normal mode\n"
        "Write another line in the same file"
    ) == rs


def test_open_file_common_gzip(target_path, encoding):
    file = Fl(
        path=target_path / "test_common_file.gz.text",
        encoding=encoding,
        compress="gzip",
    )
    with file.open(mode="w") as f:
        f.write("Write data with common file in gzip mode")

    with file.open(mode="r") as f:
        rs = f.read()

    assert "Write data with common file in gzip mode" == rs


def test_open_file_common_gz(target_path, encoding):
    file = Fl(
        path=target_path / "test_common_file.gz.text",
        encoding=encoding,
        compress="gz",
    )
    with file.open(mode="w") as f:
        f.write("Write data with common file in gz mode")

    with file.open(mode="r") as f:
        rs = f.read()

    assert "Write data with common file in gz mode" == rs


def test_open_file_common_xz(target_path, encoding):
    file = Fl(
        path=target_path / "test_common_file.xz.text",
        encoding=encoding,
        compress="xz",
    )
    with file.open(mode="w") as f:
        f.write("Write data with common file in xz mode")

    with file.open(mode="r") as f:
        rs = f.read()

    assert "Write data with common file in xz mode" == rs


def test_open_file_common_bz2(target_path, encoding):
    file = Fl(
        path=target_path / "test_common_file.bz2.text",
        encoding=encoding,
        compress="bz2",
    )
    with file.open(mode="w") as f:
        f.write("Write data with common file in bz2 mode")

    with file.open(mode="r") as f:
        assert "Write data with common file in bz2 mode" == f.read()


def test_open_file_binary(target_path, encoding):
    file = Fl(
        path=target_path / "test_binary_file.text",
        encoding=encoding,
    )
    with file.open(mode="wb") as f:
        f.write(b"Write data with binary file in normal mode")

    with file.open(mode="rb") as f:
        assert b"Write data with binary file in normal mode" == f.read()


def test_open_file_binary_gzip(target_path, encoding):
    file = Fl(
        path=target_path / "test_binary_file.gz.text",
        encoding=encoding,
        compress="gzip",
    )
    with file.open(mode="wb") as f:
        f.write(b"Write data with binary file in gzip mode")

    with file.open(mode="rb") as f:
        assert b"Write data with binary file in gzip mode" == f.read()


def test_open_file_binary_gz(target_path, encoding):
    file = Fl(
        path=target_path / "test_binary_file.gz.text",
        encoding=encoding,
        compress="gz",
    )
    with file.open(mode="wb") as f:
        f.write(b"Write data with binary file in gz mode")

    with file.open(mode="rb") as f:
        rs = f.read()

    assert b"Write data with binary file in gz mode" == rs


def test_open_file_binary_xz(target_path, encoding):
    file = Fl(
        path=target_path / "test_binary_file.xz.text",
        encoding=encoding,
        compress="xz",
    )
    with file.open(mode="wb") as f:
        f.write(b"Write data with binary file in xz mode")

    with file.open(mode="rb") as f:
        assert b"Write data with binary file in xz mode" == f.read()


def test_open_file_binary_bz2(target_path, encoding):
    file = Fl(
        path=target_path / "test_binary_file.bz2.text",
        encoding=encoding,
        compress="bz2",
    )
    with file.open(mode="wb") as f:
        f.write(b"Write data with binary file in bz2 mode")

    with file.open(mode="rb") as f:
        rs = f.read()

    assert b"Write data with binary file in bz2 mode" == rs



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import fnmatch
from collections.abc import Collection, Iterator
from pathlib import Path
from typing import Optional, Union

from .__type import Icon, icons


def replace_sep(value: str) -> str:
    return value.replace("\\", "/")


def glob_files(path: Path) -> Iterator[Path]:  # pragma: no cov
    yield from (file for file in path.rglob("*") if file.is_file())


class PathSearch:
    """Path Search object that use to search path tree from an input root path.
    It allows you to adjust recursive level value and exclude dir or file paths
    on the searching process.

    :param root: An input root path that want to search.
    :param exclude: A list of exclude paths.
    """

    def __init__(
        self,
        root: Union[str, Path],
        *,
        exclude: Optional[list[str]] = None,
        max_level: int = -1,
        length: int = 4,
        icon: int = 1,
    ) -> None:
        self.root: Path = Path(root) if isinstance(root, str) else root

        if not self.root.exists():
            raise FileNotFoundError(f"Does not found {self.root.resolve()}")

        self.exclude: list[str] = exclude or []
        self.max_level: int = max_level
        self.length: int = length
        self.real_level: int = 0

        # NOTE: Define icon argument and check an input length.
        self.icon: Icon = icons(icon)

        assert (
            len(self.icon) + 1
        ) < self.length, "a `length` argument must gather than length of icon."

        self.output_buf: list = [f"[{self.root.stem}]"]
        self.files: list[Path] = []
        self.__recurse(self.root, list(self.root.glob("*")), "", 0)

    @property
    def level(self) -> int:
        """Return level of sub path from the root path."""
        return self.real_level + 1 if self.max_level == -1 else self.max_level

    def __recurse(
        self,
        path: Path,
        file_list: list[Path],
        prefix: str,
        level: int,
    ):
        """Path recursive method for generate buffer of tree and files."""
        if not file_list or (self.max_level != -1 and self.max_level <= level):
            return

        self.real_level: int = max(level, self.real_level)
        file_list.sort(key=lambda f: (path / f).is_file())
        for i, sub_path in enumerate(file_list):

            if any(fnmatch.fnmatch(sub_path.name, exc) for exc in self.exclude):
                continue

            full_path: Path = path / sub_path
            idc: str = (
                self.icon.last if i == (len(file_list) - 1) else self.icon.next
            )

            if full_path.is_dir():
                self.output_buf.append(f"{prefix}{idc}[{sub_path}]")
                tmp_prefix: str = (
                    (
                        f"{prefix}{self.icon.normal}"
                        f'{" " * (self.length - len(self.icon))}'
                    )
                    if len(file_list) > 1 and i != len(file_list) - 1
                    else f'{prefix}{" " * self.length}'
                )
                self.__recurse(
                    full_path, list(full_path.iterdir()), tmp_prefix, level + 1
                )
            elif full_path.is_file():  # pragma: no cov
                self.output_buf.append(f"{prefix}{idc}{sub_path}")
                self.files.append(full_path)

    def pick(self, filename: Union[str, Collection[str]]) -> list[Path]:
        """Return filename with match with input argument."""
        patterns = (filename,) if isinstance(filename, str) else filename
        return list(
            filter(
                (
                    lambda f: any(
                        fnmatch.fnmatch(f, f"*/{pattern}")
                        for pattern in patterns
                    )
                ),
                self.files,
            )
        )

    def tree(self, newline: Optional[str] = None) -> str:  # pragma: no cov
        """Return path tree of root path."""
        return (newline or "\n").join(self.output_buf)


def read_ignore(file: Union[str, Path]) -> list[str]:
    """Read ignore file content to list of ignore pattern."""
    ignores: list[str] = [file.name]
    if file.exists():
        ignores.extend(
            [
                line.strip()
                for line in file.read_text().splitlines()
                if line.strip() and not line.strip().startswith("#")
            ]
        )
    return ignores


def is_ignored(file: Path, ignores: list[str]) -> bool:
    """Check if a path should be ignored based on patterns.

    :params file: (Path): a file path to check against ignore patterns
    :params ignores:

    :returns: True if path should be ignored, False otherwise
    """
    return any(
        (file.match(f"**/{pattern}/*") or file.match(f"**/{pattern}*"))
        for pattern in ignores
    )


def ls(path: Union[str, Path], ignore_file: Optional[str] = None) -> list[Path]:
    """List files in a directory, applying ignore-style filtering.

    :params base_dir: (str | Path) Base directory to search for files.
    :params ignore_file: (str) Name of the ignore file.

    :return: (list[Path]) Paths of file that are not ignored.
    """
    path: Path = Path(path).resolve()
    ignores: list[str] = read_ignore(path / ignore_file) if ignore_file else []
    return [
        file
        for file in glob_files(path)
        if not is_ignored(file, ignores=ignores)
    ]



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from dataclasses import dataclass
from typing import Union

TupleStr = tuple[str, ...]
AnyValue = Union[str, int, float, bool, None]
AnyData = Union[
    AnyValue, dict[str, AnyValue], list[AnyValue], list[dict[str, AnyValue]]
]


@dataclass(frozen=True)  # pragma: no cov
class Icon:
    """Icon dataclass object that keep necessary element for making tree."""

    normal: str
    next: str
    last: str

    def __len__(self) -> int:
        """Return the maximum length of element characters."""
        return max(len(self.normal), len(self.next), len(self.last))


def icons(theme: int) -> Icon:
    """Get the Icon object from a theme value."""
    return {
        1: Icon(normal="│", next="├─", last="└─"),
        2: Icon(normal="┃", next="┣━", last="┗━"),
        3: Icon(normal="│", next="├─", last="╰─"),
    }[theme]



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import sys
import zipfile
from pathlib import Path
from tarfile import TarFile
from typing import (
    Any,
    Literal,
    Optional,
    Protocol,
    Union,
)

from ddeutil.core import splitter

DirCompressType = Literal["zip", "rar", "tar", "h5", "hdf5", "fits"]


class OpenDirProtocol(Protocol):  # pragma: no cov
    """Open Directory Protocol object."""

    def write(self, name, arcname: Optional[str]): ...

    def safe_extract(self, path: Union[str, Path], members=None): ...


class CustomZipFl(zipfile.ZipFile):
    """Override ZipFile object."""

    def safe_extract(self, path: str, members=None):
        self.extractall(path, members)


class CustomTarFl(TarFile):
    """Override TarFile object."""

    def write(self, name, arcname: Optional[str] = None):
        """Clone ``self.add`` method to the new name."""
        return self.add(name, arcname)

    def safe_extract(self, path: Union[str, Path] = ".", members=None):
        path: Path = path if isinstance(path, Path) else Path(path)

        # NOTE: For Python version >= 3.12
        if sys.version_info >= (3, 12):  # pragma: no cov
            self.extractall(path, members, filter="data")
            return

        self.extractall(path, members)


class Dir:
    """Open Dir Object."""

    def __init__(
        self,
        path: Union[str, Path],
        *,
        compress: str,
    ) -> None:
        self.path: Path = Path(path) if isinstance(path, str) else path

        # NOTE: Split compress and sub-compress
        _compress, sub = splitter.must_split(compress, ":", maxsplit=1)
        self.compress: str = _compress
        self.sub_compress: str = sub or "_"

        # NOTE: Action anything after set up attributes.
        self.after_set_attrs()

    def after_set_attrs(self) -> None: ...

    def open(
        self,
        *,
        mode: Literal["r", "w", "x", "a"],
        **kwargs,
    ) -> OpenDirProtocol:
        """Open dir"""
        if self.compress in ("zip",):
            zip_compress: dict[str, Any] = {
                "_": zipfile.ZIP_STORED,
                "zlib": zipfile.ZIP_DEFLATED,
                "bz2": zipfile.ZIP_BZIP2,
                "lzma": zipfile.ZIP_LZMA,
            }

            return CustomZipFl(
                self.path,
                mode=mode,
                compression=zip_compress[self.sub_compress],
                **kwargs,
            )
        elif self.compress in ("tar",):
            tar_compress: dict[str, str] = {
                "_": "gz",
                "gz": "gz",
                "bz2": "bz2",
                "xz": "xz",
            }

            return CustomTarFl.open(
                self.path,
                mode=f"{mode}:{tar_compress[self.sub_compress]}",
            )
        raise NotImplementedError



import os
from dataclasses import asdict
from pathlib import Path
from textwrap import dedent

import pytest
from ddeutil.io.config import Params, Paths, Rule, Stage
from ddeutil.io.exceptions import ConfigArgumentError


@pytest.fixture(scope="module")
def toml_conf_path(test_path):
    with open(test_path / "io-register.toml", mode="w") as f:
        f.write(
            dedent(
                """
                [tool.io.register.paths]
                root = "./"
                data = "./data"
                conf = "./conf"

                [tool.io.register.stages]
                raw = {format = "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"}
                persisted = {format = "{naming:%s}.{version:v%m.%n.%c}"}
                """.strip(
                    "\n"
                )
            )
        )
    yield test_path / "io-register.toml"
    os.unlink(test_path / "io-register.toml")


@pytest.fixture(scope="module")
def yaml_conf_path(test_path):
    with open(test_path / "io-register.yaml", mode="w") as f:
        f.write(
            dedent(
                """
            tool:
                io:
                    register:
                        paths:
                            root: "./"
                            data: "./data"
                            conf: "./conf"
                        stages:
                            raw:
                                layer: 1
                                format: "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"
                            persisted:
                                layer: 2
                                format: "{naming:%s}.{version:v%m.%n.%c}"
                """.strip(
                    "\n"
                )
            )
        )
    yield test_path / "io-register.yaml"
    os.unlink(test_path / "io-register.yaml")


def test_config_rule():
    assert {
        "timestamp": {},
        "excluded": [],
        "compress": None,
    } == asdict(Rule())


def test_config_stage():
    stage: Stage = Stage(
        **{
            "alias": "persisted",
            "format": "{timestamp:%Y-%m-%d}{naming:%c}.json",
            "rule": {
                "timestamp": {"minutes": 15},
            },
        }
    )
    assert stage.rule.compress is None
    assert stage.alias == "persisted"
    assert stage.format == "{timestamp:%Y-%m-%d}{naming:%c}.json"


def test_config_stage_raise():
    with pytest.raises(ConfigArgumentError):
        Stage(
            **{
                "alias": "persisted",
                "format": "{ts:%Y-%m-%d}{naming:%c}.json",
            }
        )

    with pytest.raises(ConfigArgumentError):
        Stage(
            **{
                "alias": "persisted",
                "format": "fix_format_name.json",
            }
        )

    with pytest.raises(ConfigArgumentError):
        Stage(
            **{
                "alias": "persisted",
                "format": "{naming:%c}.json",
                "rule": {
                    "timestamp": {"minutes": 15},
                },
            }
        )


def test_config_paths():
    paths = Paths()
    assert Path() == paths.root
    assert Path() / "data" == paths.data
    assert Path() / "conf" == paths.conf

    os.environ["IO_REGISTER_ROOT_PATH"] = "./core"
    paths = Paths()
    assert Path("./core") == paths.root

    os.environ.pop("IO_REGISTER_ROOT_PATH")

    paths = Paths(
        root="./core",
        data="./core/data",
        conf="../conf",
    )
    assert Path("./core") == paths.root
    assert Path("./core/data") == paths.data
    assert Path("../conf") == paths.conf


def test_config_params():
    params: Params = Params(
        **{
            "stages": {
                "raw": {"format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"},
                "persisted": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
                "curated": Stage(
                    **{"format": "{domain:%s}_{naming:%s}.{compress:%-g}"}
                ),
            }
        }
    )
    assert {
        "stages": {
            "raw": {
                "format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}",
                "alias": "raw",
                "rule": {
                    "timestamp": {},
                    "excluded": [],
                    "compress": None,
                },
                "layer": 1,
            },
            "persisted": {
                "format": "{naming:%s}.{version:v%m.%n.%c}",
                "alias": "persisted",
                "rule": {
                    "timestamp": {},
                    "excluded": [],
                    "compress": None,
                },
                "layer": 2,
            },
            "curated": {
                "format": "{domain:%s}_{naming:%s}.{compress:%-g}",
                "alias": "curated",
                "rule": {
                    "timestamp": {},
                    "excluded": [],
                    "compress": None,
                },
                "layer": 3,
            },
        },
        "paths": {
            "conf": Path() / "./conf",
            "data": Path() / "./data",
            "root": Path(),
        },
    } == asdict(params)

    assert params.stage_first == "raw"
    assert params.stage_final == "curated"
    assert params.get_stage(name="base").alias == "base"

    with pytest.raises(ConfigArgumentError):
        params.get_stage("not_found")


def test_config_params_raise():
    with pytest.raises(ConfigArgumentError):
        Params(**{"stages": {"raw": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"}})


def test_config_params_toml(toml_conf_path):
    params = Params.from_toml(toml_conf_path)
    assert {
        "stages": {
            "raw": {
                "format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}",
                "alias": "raw",
                "rule": {
                    "timestamp": {},
                    "excluded": [],
                    "compress": None,
                },
                "layer": 1,
            },
            "persisted": {
                "format": "{naming:%s}.{version:v%m.%n.%c}",
                "alias": "persisted",
                "rule": {
                    "timestamp": {},
                    "excluded": [],
                    "compress": None,
                },
                "layer": 2,
            },
        },
        "paths": {
            "conf": Path("conf"),
            "data": Path("data"),
            "root": Path("."),
        },
    } == params.to_dict()


def test_config_params_yaml(yaml_conf_path):
    params = Params.from_yaml(yaml_conf_path)
    assert {
        "stages": {
            "raw": {
                "format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}",
                "alias": "raw",
                "rule": {
                    "timestamp": {},
                    "excluded": [],
                    "compress": None,
                },
                "layer": 1,
            },
            "persisted": {
                "format": "{naming:%s}.{version:v%m.%n.%c}",
                "alias": "persisted",
                "rule": {
                    "timestamp": {},
                    "excluded": [],
                    "compress": None,
                },
                "layer": 2,
            },
        },
        "paths": {
            "conf": Path("conf"),
            "data": Path("data"),
            "root": Path("."),
        },
    } == asdict(params)



import shutil
import time
from collections.abc import Iterator
from pathlib import Path

import pytest
import yaml
from ddeutil.io.config import Params
from ddeutil.io.exceptions import RegisterArgumentError
from ddeutil.io.register import ArchiveRegister, Register


@pytest.fixture(scope="module")
def target_path(test_path) -> Iterator[Path]:
    tgt_path: Path = test_path / "register_archive_temp"
    tgt_path.mkdir(exist_ok=True)
    (tgt_path / "conf/demo").mkdir(parents=True, exist_ok=True)
    with open(tgt_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "conn.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/dummy",
                }
            },
            f,
        )
    yield tgt_path

    shutil.rmtree(tgt_path)


@pytest.fixture(scope="module")
def params(target_path, root_path) -> Params:
    return Params(
        **{
            "paths": {"root": target_path},
            "stages": {
                "raw": {
                    "format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}",
                    "rule": {"timestamp": {"seconds": 2}},
                },
                "persisted": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
            },
        }
    )


@pytest.fixture(scope="module")
def params_archive(target_path, root_path) -> Params:
    return Params(
        **{
            "paths": {
                "root": target_path,
                "data": target_path / "data_archive",
            },
            "stages": {
                "raw": {
                    "format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}",
                    "rule": {"timestamp": {"seconds": 2}},
                },
                "persisted": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
            },
        }
    )


def test_register_deploy(params, target_path):
    Register(name="demo:conn_local_file", params=params).deploy()
    with open(target_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/new",
                    "extra": {"foo": "bar"},
                }
            },
            f,
        )

    time.sleep(1)

    register = Register(name="demo:conn_local_file", params=params).deploy()
    assert len(list((target_path / "data/raw").rglob("*"))) == 2

    time.sleep(2)

    register.purge(stage="persisted")

    with open(target_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/update",
                    "extra": {"foo": "baz"},
                }
            },
            f,
        )

    Register(name="demo:conn_local_file", params=params).deploy()
    assert len(list((target_path / "data/raw").rglob("*"))) != 3


def test_register_deploy_raise(params, target_path):
    with pytest.raises(RegisterArgumentError):
        Register(name="demo:conn_local_file", params=params).deploy(stop="foo")

    with pytest.raises(RegisterArgumentError):
        Register(name="demo:conn_local_file", params=params).remove()


def test_register_archive_deploy(params_archive, target_path):
    ArchiveRegister(name="demo:conn_local_file", params=params_archive).deploy()
    with open(target_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/new",
                    "extra": {"foo": "bar"},
                }
            },
            f,
        )

    time.sleep(1)
    register = ArchiveRegister(
        name="demo:conn_local_file", params=params_archive
    ).deploy()
    assert len(list((target_path / "data_archive/raw").rglob("*"))) == 2

    time.sleep(2)

    register.purge(stage="persisted")

    with open(target_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/update",
                    "extra": {"foo": "baz"},
                }
            },
            f,
        )

    ArchiveRegister(name="demo:conn_local_file", params=params_archive).deploy()
    assert len(list((target_path / "data_archive/raw").rglob("*"))) != 3

    assert (
        len(
            list(
                (
                    target_path / f"data_archive/{ArchiveRegister.archiving}"
                ).rglob("*")
            )
        )
        > 0
    )

    register_persisted = ArchiveRegister(
        name="demo:conn_local_file", params=params_archive
    ).switch("persisted")

    register_persisted.remove()


def test_register_archive_deploy_raise(params_archive, target_path):
    with pytest.raises(RegisterArgumentError):
        ArchiveRegister(
            name="demo:conn_local_file", params=params_archive
        ).remove()



import csv
import os
import shutil
from collections.abc import Iterator
from pathlib import Path

import pytest
from ddeutil.io.utils import (
    map_func,
    reverse_readline,
    search_env,
    search_env_replace,
    template_func,
    template_secret,
)


def test_template_secret():
    assert "Value include secrets: s3://bar" == template_secret(
        "Value include secrets: s3://@secrets{foo}",
        secrets={"foo": "bar"},
    )

    rs = template_secret(
        {
            "list": ["1", "2", "s3://@secrets{foo}"],
            "dict": {
                "tuple": ("1", "2", "s3://@secrets{foo}"),
                "key": 1,
                "boolean": True,
            },
            "default": "s3://@secrets{test:default}",
        },
        secrets={"foo": "bar"},
    )
    assert {
        "list": ["1", "2", "s3://bar"],
        "dict": {
            "tuple": ("1", "2", "s3://bar"),
            "key": 1,
            "boolean": True,
        },
        "default": "s3://default",
    } == rs


def test_template_secret_raise():
    with pytest.raises(ValueError):
        template_secret(
            "Value include secrets: s3://@secrets{foo.name}",
            secrets={"foo": "bar"},
        )


def test_template_func():
    assert "Test a|" == template_func(
        "Test @function{ddeutil.io.utils.add_newline:'a',newline='|'}"
    )

    reuse: str = "@function{ddeutil.io.utils.add_newline:'a',newline='|'}"
    assert {
        "list": ["a|", 1],
        "tuple": ("a|", 2, 3),
    } == template_func(
        {
            "list": [reuse, 1],
            "tuple": (reuse, 2, 3),
        }
    )


def test_template_func_raise():
    with pytest.raises(ValueError):
        template_func("@function{ddeutil.io.__version__:'a'}")


def test_map_func():
    assert {"foo": "bar!"} == map_func({"foo": "bar"}, lambda x: x + "!")
    assert ("foo!", "bar!", 1) == map_func(("foo", "bar", 1), lambda x: x + "!")


@pytest.fixture(scope="module")
def utils_path(test_path) -> Iterator[Path]:
    this_path: Path = test_path / "utils_reverse"
    this_path.mkdir(parents=True, exist_ok=True)

    yield this_path

    shutil.rmtree(this_path)


@pytest.fixture(scope="module")
def csv_data() -> list[dict[str, str]]:
    return [
        {"Col01": "A", "Col02": "1", "Col03": "test1"},
        {"Col01": "B", "Col02": "2", "Col03": "test2"},
        {"Col01": "C", "Col02": "3", "Col03": "test3"},
    ]


def test_files_utils_search_env_replace():
    os.environ["NAME"] = "foo"
    assert "Hello foo" == search_env_replace("Hello ${NAME}")


def test_files_utils_search_env_replace_raise():
    with pytest.raises(ValueError):
        search_env_replace(
            "Hello ${NAME01}",
            raise_if_default_not_exists=True,
        )

    with pytest.raises(ValueError):
        search_env_replace("Hello ${:test}")


def test_files_utils_search_env():
    assert {
        "key": "demo",
        "hello": "demo-2",
        "escape": "${key}",
    } == search_env(
        "key='demo'\n# foo=bar\nhello=${key}-2\nescape=\\${key}\n",
    )


def test_files_utils_search_env_raise():
    with pytest.raises(ValueError):
        search_env("foo=")

    with pytest.raises(ValueError):
        search_env("foo=''")

    with pytest.raises(ValueError):
        search_env('foo=""')


def test_files_utils_reverse_read(utils_path, csv_data):
    test_file = utils_path / "file_reverse.csv"

    with open(test_file, mode="w", newline="") as f:
        writer = csv.DictWriter(
            f,
            fieldnames=list(csv_data[0].keys()),
            lineterminator="\n",
        )
        writer.writerows(csv_data)

    with open(test_file) as f:
        rs = list(reverse_readline(f))
    assert rs == [
        "C,3,test3\n",
        "B,2,test2\n",
        "A,1,test1\n",
    ]



import json
import shutil
from collections.abc import Generator
from datetime import datetime
from pathlib import Path
from unittest.mock import patch

import pytest
import yaml
from ddeutil.io.config import Params
from ddeutil.io.register import Register


@pytest.fixture(scope="module")
def target_path(test_path) -> Generator[Path, None, None]:
    tgt_path: Path = test_path / "register_deploy_temp"
    tgt_path.mkdir(exist_ok=True)
    (tgt_path / "conf/demo").mkdir(parents=True, exist_ok=True)
    with open(tgt_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "conn.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/dummy",
                }
            },
            f,
        )
    yield tgt_path
    shutil.rmtree(tgt_path)


@pytest.fixture(scope="module")
def params(target_path, root_path) -> Params:
    return Params(
        **{
            "paths": {"root": target_path},
            "stages": {
                "raw": {"format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"},
                "staging": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
                "persisted": {
                    "format": "{domain:%s}_{naming:%s}.{compress:%-g}",
                    "rule": {"compress": "gzip"},
                },
            },
        }
    )


@pytest.fixture(scope="module")
def mock_get_date():
    with patch(
        target="ddeutil.io.register.get_date",
        return_value=datetime(2024, 1, 1, 1),
    ) as mock:
        yield mock


def test_register_deploy(params, target_path, mock_get_date):
    assert mock_get_date.mocked

    data_path = target_path / "data"
    Register(name="demo:conn_local_file", params=params).deploy()
    assert (data_path / "staging/conn_local_file.v0.0.1.json").exists()
    assert (data_path / "raw/conn_local_file.20240101_010000.json").exists()
    assert (data_path / "persisted/demo_conn_local_file.gz.json").exists()


def test_register_multiple_files(params, target_path, root_path):
    data_path = target_path / "data"
    Register(name="demo:conn_local_file", params=params).deploy()

    with open(data_path / "raw/conn_local_file_new.json", mode="w") as f:
        json.dump({"foo": "bar"}, f)

    register = Register(name="demo:conn_local_file", stage="raw", params=params)
    assert str(register.version()) == "0.0.1"



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
"""Register module provide the Register object for manage configuration file
with multiple storing stages. That is mean if you have original config file with
YAML file format and want to save memory to reading in your any application
you will convert it to JSON file or some binary file and always reuse it if the
original config file does not change context data inside.

    So, this module will help you handle this scenario with Register object.
This object can dynamic stage with your params config.
"""
from __future__ import annotations

import logging
import os
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Optional, TypedDict
from zoneinfo import ZoneInfo

from ddeutil.core import base, hash, merge, splitter
from ddeutil.core.dtutils import get_date

try:
    from deepdiff import DeepDiff
except ImportError as err:  # pragma: no cov
    raise ImportError(
        "Register module need `deepdiff` package, so, please install it by "
        "`pip install deepdiff`"
    ) from err

try:
    from fmtutil import (
        ConstantType,
        Datetime,
        FormatterArgumentError,
        FormatterGroup,
        FormatterGroupType,
        Naming,
        VerPackage,
        Version,
        make_const,
        make_group,
    )
except ImportError as err:  # pragma: no cov
    raise ImportError(
        "Register module need `fmtutil` package, so, please install it by "
        "`pip install fmtutil`"
    ) from err

try:
    from dateutil.relativedelta import relativedelta
except ImportError:  # pragma: no cov
    relativedelta = None


if TYPE_CHECKING:
    from typing_extensions import Self

from .__type import AnyData, TupleStr
from .config import (
    DATE_FMT,
    DATE_LOG_FMT,
    UPDATE_KEY,
    VERSION_KEY,
    Params,
    Rule,
)
from .exceptions import RegisterArgumentError, StoreNotFound
from .stores import Store
from .utils import rm

logger = logging.getLogger("ddeutil.io")

REGISTER_BASE_STAGE_DEFAULT: str = "base"
REGISTER_DIFF_LEVEL: dict[int, TupleStr] = {
    1: (
        "values_changed",
        "type_changes",
    ),
    2: (
        "dictionary_item_added",
        "dictionary_item_removed",
        "iterable_item_added",
        "iterable_item_removed",
    ),
}

__all__: TupleStr = (
    "Register",
    "ArchiveRegister",
)


class StageFl(TypedDict):
    """Stage files dict typing for the mypy checker step."""

    parse: FormatterGroup
    file: str


CompressConst: ConstantType = make_const(
    name="CompressConst",
    formatter={
        "%g": "gzip",
        "%-g": "gz",
        "%b": "bz2",
        "%r": "rar",
        "%x": "xz",
        "%z": "zip",
    },
)

FileExtensionConst: ConstantType = make_const(
    name="FileExtensionConst",
    formatter={
        "%j": "json",
        "%y": "yaml",
        "%e": "env",
        "%t": "toml",
    },
)


class BaseRegister:
    """Base Register object that is not implement any features without base
    properties.

    :param name: A name of key of config data that want to register.
    :param domain: A dir path of config files that use to search a name of data.
    """

    metadata: ClassVar[str] = "__METADATA"

    def __init__(self, name: str, *, domain: Optional[str] = None) -> None:
        self.name: str = name
        self.domain: str = (
            domain.replace(os.sep, "/").strip("/").lower() if domain else ""
        )
        if any(sep in self.name for sep in (",", ".")):
            raise RegisterArgumentError(
                "The register name should not contain any `,` or `.` "
                "characters."
            )

        # NOTE: generate update datetime.
        self.updt: datetime = get_date("datetime").replace(microsecond=0)

    @property
    def fullname(self) -> str:
        """Return a configuration fullname, which join `name` and `domain`
        together with domain partition string.

        :rtype: str
        """
        return f"{self.domain}:{self.name}" if self.domain else self.name

    @property
    def shortname(self) -> str:
        """Return a configuration shortname, which get first character of any
        split string by name partition string.

        :rtype: str
        """
        return base.concat(word[0] for word in self.name.split("_"))

    @property
    def fmt_type(self) -> FormatterGroupType:  # pragma: no cov
        """Return the generated formatter group that include constant formatters
        from ``self.name`` and ``self.domain``.

        :rtype: FormatterGroupType
        """
        return make_group(
            {
                "naming": make_const(fmt=Naming, value=self.name),
                "domain": make_const(fmt=Naming, value=self.domain),
                "compress": CompressConst,
                "extension": FileExtensionConst,
                "version": Version,
                "timestamp": Datetime,
            }
        )


class Register(BaseRegister):
    """Register Object that contain configuration config methods and metadata
    management. This object work with stage input argument, that set all
    properties in the `parameter.yaml` file.

    :param name: A fullname of register data that able to split domain and name
        before passing to parent base register object with ``:``.
    :param stage: A stage name that want to get data with an input name.
    :param params:
    :param store:
    """

    @classmethod
    def reset(
        cls,
        name: str,
        params: Params,
    ) -> Self:
        """Reset all configuration data files that exists in any stage but
        does not do anything in the base stage. This method will use when the
        config name of data was changed and does not use the old name. If the
        name was changed and that config data does not reset,
        the configuration files of this data will exist in any moved stage.

        :param name: The fullname of configuration.
        :type name: str
        :param params:
        :type params: Params

        :rtype: Self
        :return: itself object that passing the fullname to initialize step.
        """
        domain, n = cls.split_domain(name)
        for stage in params.stages:
            try:
                # NOTE: Start reset (remove) on the target stage area.
                cls(name, stage=stage, params=params).remove()

                # NOTE: Remove the metadata file on this stage.
                meta_file: Path = (
                    params.paths.data
                    / cls.metadata
                    / f"{domain or ''}{n}.{stage}.json"
                )
                rm(meta_file)
            except StoreNotFound:
                continue

        # NOTE: Remove the metadata file on the base stage.
        meta_file: Path = (
            params.paths.data / cls.metadata / f"{domain or ''}{n}.base.json"
        )
        rm(meta_file, force_raise=False)
        return cls(name, params=params)

    @classmethod
    def split_domain(cls, name: str) -> tuple[str, str]:
        rs: list[str] = splitter.must_rsplit(
            base.concat(name.split()),
            sep=":",
            maxsplit=1,
        )
        return rs[0], rs[-1]

    def __init__(
        self,
        name: str,
        stage: Optional[str] = None,
        *,
        params: Optional[Params] = None,
        store: Optional[type[Store]] = None,
    ) -> None:
        domain, name = self.split_domain(name)
        super().__init__(name=name, domain=domain)
        if not params:
            raise NotImplementedError(
                "This register instance can not do any actions because config "
                "param does not set."
            )
        self.params: Optional[Params] = params
        self.stage: str = stage or REGISTER_BASE_STAGE_DEFAULT
        self.store: Optional[type[Store]] = store

        # NOTE: Load latest version of data from data lake or data store of
        #   configuration files.
        self.__raw_data: AnyData = self.get(stage=self.stage)
        if not self.__raw_data:
            raise StoreNotFound(
                f"Register name {self.name!r} "
                f"{f'in domain {self.domain!r} ' if self.domain else ' '}"
                f"does not find data in stage: {self.stage!r}."
            )

        # NOTE: Running metadata tracking cache.
        self.meta: dict[str, Any] = {}
        self.__manage_metadata()

    def __manage_metadata(self):
        """Manage the latest context data for detect change by metadata
        strategy.

            The metadata file also keeps on ./data/<self.metadata> dir for all
        config context data.
        """
        store: Store = Store(path=self.params.paths.data / self.metadata)
        meta_file: Path = (
            store.path / f"{self.domain or ''}{self.name}.{self.stage}.json"
        )
        self.meta: dict[str, Any] = store.load(path=meta_file, default={})

        # NOTE: Compare data from current stage and latest version in metadata.
        self.changed: int = self.compare_data(self.meta)

        # NOTE: Update metadata if the configuration data does not exist, or it
        #   has any changes.
        if self.changed > 0:
            if self.changed == 99:
                logger.debug(
                    f"Data in stage: {self.stage!r} does not exists in metadata"
                )
            else:
                logger.debug(
                    f"Should update metadata because diff level is "
                    f"{self.changed}."
                )
            store.save(path=meta_file, data=self.data(hashing=True))

    def __str__(self) -> str:
        return f"({self.fullname}, {self.stage})"

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}(name={self.fullname!r}"
            f"{f'stage={self.stage!r}' if self.stage != 'base' else ''})>"
        )

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.fullname == other.fullname
                and self.stage == other.stage
                and self.timestamp == other.timestamp
            )
        return NotImplemented

    def data(self, hashing: bool = False) -> dict[str, Any]:
        """Return the context data with the specific name of data.

        :param hashing: A hashing flag that allow use hash function on the
            context data.
        :type hashing: bool (Default=False)
        :rtype: dict[str, Any]
        """
        _data: dict[str, Any] = self.__raw_data.copy()
        if not self.stage or (self.stage == REGISTER_BASE_STAGE_DEFAULT):
            _data: dict[str, Any] = {
                k: self.meta[k]
                for k in self.meta
                if k in (UPDATE_KEY, VERSION_KEY)
            } | self.__raw_data

        return (
            hash.hash_value(_data, exclude={UPDATE_KEY, VERSION_KEY})
            if hashing
            else _data
        )

    @property
    def timestamp(self) -> datetime:
        """Return the current timestamp value of config data. If timestamp value
        does not exist. this property will return timestamp of initialize.

        :rtype: datetime
        """
        if self.changed > 0:
            return self.updt
        elif dt := self.data().get(UPDATE_KEY):
            return datetime.strptime(dt, DATE_FMT).replace(
                tzinfo=ZoneInfo("UTC")
            )
        return self.updt

    def version(self) -> VerPackage:
        """Generate version value from the get method. If version value does
        not exist from configuration data, this property will return the
        default, `v0.0.1`. If the initialization process tracking some change
        from configuration data between metadata and the latest data in the
        stage, the _next will be generated.

        :rtype: VerPackage
        """
        version = VerPackage.parse(self.data().get(VERSION_KEY, "v0.0.1"))
        if self.changed == 0 or self.changed == 99:
            return version
        elif self.changed == 1:
            return version.bump_patch()
        elif self.changed == 2:
            return version.bump_minor()
        else:  # pragma: no cov
            return version.bump_major()

    def fmt(self, update: Optional[dict[str, Any]] = None) -> FormatterGroup:
        """Return FormatterGroup object that passing ``self.timestamp`` and
        ``self.version`` values.

        :param update: An update values on the formatter group object.

        :rtype: FormatterGroup
        """
        return self.fmt_type(
            {
                "timestamp": self.timestamp,
                "version": self.version(),
                **(update or {}),
            }
        )

    def compare_data(self, data: dict[str, Any]) -> int:
        """Return difference column from dictionary comparison method which use
        the `deepdiff` library.

        :param data: dict : The data dictionary for compare with current
            configuration data.

        :rtype: Literal[0, 1, 2, 99]
        """
        if not data:
            return 99

        rs: DeepDiff = DeepDiff(
            self.data(hashing=True),
            data,
            ignore_order=True,
            exclude_paths=[f"root[{k!r}]" for k in (UPDATE_KEY, VERSION_KEY)],
        )
        if not rs:
            return 0
        for level, diffs in REGISTER_DIFF_LEVEL.items():
            if any(_ in rs for _ in diffs):
                return level
        raise NotImplementedError(  # pragma: no cov
            f"Compare change do not implement for deepdiff result: {rs}"
        )

    def _stage_files(
        self,
        stage: str,
        store: Store,
    ) -> dict[int, StageFl]:
        """Return the mapping of StageFl data from target stage area.

        :param stage: A stage value that want to search files.
        :param store: A store object that passing path with stage path.

        :rtype: dict[int, StageFl]
        """
        rs: dict[int, StageFl] = {}
        for index, file in enumerate((_f.name for _f in store.ls()), start=1):
            try:
                rs[index]: StageFl = {
                    "parse": self.fmt_type.parse(
                        value=file,
                        fmt=rf"{self.params.get_stage(stage).format}\.json",
                    ),
                    "file": file,
                }
            except FormatterArgumentError:
                continue
        return rs

    def get(
        self,
        stage: Optional[str] = None,
        *,
        order: Optional[int] = 1,
        reverse: bool = False,
    ) -> AnyData:
        """Get the context data from the specific stage value (use 'base' if the
        stage do not pass on this method).

        :param stage: A stage value that want to get context data.
        :param order:
        :param reverse: A reverse flag that use to get stage file.

        :rtype: AnyData
        """
        if (stage is None) or (stage == REGISTER_BASE_STAGE_DEFAULT):
            return Store(path=(self.params.paths.conf / self.domain)).get(
                name=self.name, order=order
            )

        store: Store = Store(
            path=self.params.paths.data / stage,
            compress=self.params.get_stage(stage).rule.compress,
        )

        if rs := self._stage_files(stage, store):
            max_data: list = sorted(
                rs.items(),
                key=lambda x: (x[1]["parse"],),
                reverse=reverse,
            )
            return store.load(path=(store.path / max_data[-order][1]["file"]))
        return {}

    def move(
        self,
        stage: str,
        *,
        force: bool = False,
        retention: bool = True,
    ) -> Self:
        """Move the config data file to the target stage.

        :param stage:
        :param force:
        :param retention:

        :rtype: Self
        """
        store: Store = Store(
            path=self.params.paths.data / stage,
            compress=self.params.get_stage(stage).rule.compress,
        )
        if (
            self.compare_data(
                hash.hash_value(
                    self.get(stage=stage), exclude=(UPDATE_KEY, VERSION_KEY)
                )
            )
            > 0
            or force
        ):
            _filename: str = self.fmt().format(
                f"{self.params.get_stage(name=stage).format}.json",
            )
            if (store.path / _filename).exists():
                logger.warning(
                    f"File {_filename!r} already exists in {stage!r} stage."
                )
            store.save(
                path=(store.path / _filename),
                data=merge.merge_dict(
                    self.data(),
                    {
                        UPDATE_KEY: f"{self.timestamp:{DATE_FMT}}",
                        VERSION_KEY: f"v{str(self.version())}",
                    },
                ),
            )
            # NOTE: Retention process after move data to the stage successful
            if retention:
                self.purge(stage=stage)
        else:
            logger.warning(
                f"Config {self.name!r} cannot move {self.stage!r} -> "
                f"{stage!r} cause the data does not has any change or "
                f"force moving flag does not set."
            )
        return self.switch(stage=stage)

    def switch(self, stage: str) -> Self:
        """Switch instance from old stage to new stage with input argument.

        :param stage: A stage name that want to switch.
        """
        return self.__class__(self.fullname, stage=stage, params=self.params)

    def purge(self, stage: Optional[str] = None) -> None:
        """Purge configuration files that match with any rules in the stage
        setting.
        """
        stage: str = stage or self.stage
        rule: Rule = self.params.get_stage(stage).rule

        store: Store = Store(
            path=self.params.paths.data / stage, compress=rule.compress
        )

        if ts_timedelta := rule.timestamp:
            if relativedelta is None:  # pragma: no cov
                raise ImportError(
                    "The purge method need to install `python-dateutil` "
                    "package before."
                )

            rs: dict[int, StageFl] = self._stage_files(stage, store)

            max_index, max_stage_file = max(
                rs.items(), key=lambda x: x[1]["parse"]
            )
            rs.pop(max_index)
            upper_bound: FormatterGroup = max_stage_file["parse"].adjust(
                {"timestamp": -relativedelta(**ts_timedelta)}
            )

            for data in filter(lambda x: x["parse"] < upper_bound, rs.values()):
                print(f"Start remove {data['file']}")
                rm(store.path / data["file"])

    def deploy(self, stop: Optional[str] = None) -> Self:
        """Deploy the config data from the current stage to the final stage or
        specific an input stop stage.

        :param stop: A stage name for stop when move store from base stage
            to final stage.
        :type stop: str

        :raise RegisterArgumentError: If an input stop value does not exists on
            the stages' configuration.

        :rtype: Self
        """
        _base: Self = self
        _stop: str = stop or self.params.stage_final
        if _stop not in self.params.stages:
            raise RegisterArgumentError(
                "The `stop` argument should exists in the stages data on the "
                "Params argument."
            )

        for stage in self.params.stages:
            _base: Register = _base.move(stage)
            if stage == _stop:
                break
        return _base  # pragma: no cov

    def remove(self) -> None:
        """Remove all config files that move from the base stage by an input
        stage value. So, this method does not allow to remove data on the base
        stage.

        :raise RegisterArgumentError: If current stage be the base stage.
            Because it does not do anything on origin config files.

        :rtype: NoReturn
        """
        if self.stage == REGISTER_BASE_STAGE_DEFAULT:
            raise RegisterArgumentError(
                "The remove method can not process with the 'base' stage."
            )

        store: Store = Store(path=self.params.paths.data / self.stage)
        for stage_file in self._stage_files(self.stage, store).values():
            rm(store.path / stage_file["file"])


class ArchiveRegister(Register):
    """Archiving Register object that implement archiving management on the
    Register object such as ``self.purge``, and ``self.remove`` methods.
    """

    archiving: ClassVar[str] = ".archive"

    def purge(self, stage: Optional[str] = None) -> None:
        """Purge configuration files that match with any rules in the stage
        setting and move it to archiving area.

        :param stage: a stage value that want to purge.
        :type stage: str | None
        """
        stage: str = stage or self.stage
        rule: Rule = self.params.get_stage(stage).rule

        store: Store = Store(
            path=self.params.paths.data / stage, compress=rule.compress
        )

        if ts_timedelta := rule.timestamp:
            if relativedelta is None:  # pragma: no cov
                raise ImportError(
                    "The purge method need to install `python-dateutil` "
                    "package before."
                )

            rs: dict[int, StageFl] = self._stage_files(stage, store)
            max_index, max_stage_file = max(
                rs.items(), key=lambda x: x[1]["parse"]
            )
            rs.pop(max_index)
            upper_bound: FormatterGroup = max_stage_file["parse"].adjust(
                {"timestamp": -relativedelta(**ts_timedelta)}
            )

            for sf in filter(lambda x: x["parse"] < upper_bound, rs.values()):
                file: str = sf["file"]
                _ac_path: str = f"{stage}_{self.updt:{DATE_LOG_FMT}}_{file}"
                store.move(
                    file,
                    dest=self.params.paths.data / self.archiving / _ac_path,
                )
                rm(store.path / file)

    def remove(self) -> None:
        """Remove all config files from an input stage store area and move it to
        archiving area.

        :raise RegisterArgumentError: If current stage be the base stage.
            Because it does not do anything on origin config files.

        :rtype: NoReturn
        """
        if self.stage == REGISTER_BASE_STAGE_DEFAULT:
            raise RegisterArgumentError(
                "The remove method can not process with the 'base' stage."
            )

        store: Store = Store(path=self.params.paths.data / self.stage)

        for stage_file in self._stage_files(self.stage, store).values():
            file: str = stage_file["file"]
            store.move(
                file,
                dest=(
                    self.params.paths.data
                    / self.archiving
                    / f"{self.stage}_{self.updt:{DATE_LOG_FMT}}_{file}"
                ),
            )
            rm(store.path / file)



from __future__ import annotations

import json
import os
import shutil
from collections.abc import Iterator
from pathlib import Path

import pytest
import yaml
from ddeutil.io.stores import Store, StoreJsonToCsv, StoreToJsonLine


@pytest.fixture(scope="module")
def target_path(test_path) -> Iterator[Path]:
    """Create ./conf_file_temp/test_01_conn.yaml file on the current test path.
    This file already add data 'conn_local_file' for the store object able to
    test getting and moving.
    """
    tgt_path: Path = test_path / "store_file"
    tgt_path.mkdir(parents=True, exist_ok=True)
    with open(tgt_path / "test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/dummy",
                }
            },
            f,
        )

    with open(tgt_path / "test_01_conn.json", mode="w") as f:
        # noinspection PyTypeChecker
        json.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/dummy",
                }
            },
            f,
        )
    yield tgt_path

    shutil.rmtree(tgt_path)


@pytest.fixture(scope="module")
def new_path(test_path: Path) -> Iterator[Path]:
    new_path = test_path / "store_file_new"
    yield new_path
    shutil.rmtree(new_path)


def test_store_init(new_path):
    store = Store(new_path)
    assert new_path == store.path
    assert store.path.exists()

    store.create(new_path / "touch.json", initial_data={"foo": "bar"})
    store.create(new_path / "touch.json")


def test_store_get(target_path):
    store = Store(target_path)

    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == store.get(name="conn_local_file")

    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == store.get(name="conn_local_file", order=1)

    assert {} == store.get(name="conn_local_file_not_found")
    assert {} == store.get(name="conn_local_file", order=2)
    assert {} == store.get(name="conn_local_file", order=10)


def test_store_move(target_path):
    store = Store(target_path)
    store.move(
        "test_01_conn.yaml",
        dest=target_path / "connections/test_01_conn_new.yaml",
    )
    assert (target_path / "connections/test_01_conn_new.yaml").exists()

    base_store_temp = Store(target_path)
    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == base_store_temp.get(name="conn_local_file", order=1)
    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == base_store_temp.get(name="conn_local_file", order=2)


def test_store_json(target_path):
    store = StoreJsonToCsv(path=target_path)
    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == store.get(name="conn_local_file")


def test_store_csv_stage(target_path):
    store = StoreJsonToCsv(path=target_path)
    store.move(
        path="test_01_conn.json",
        dest=target_path / "connections/test_01_conn.move.json",
    )

    stage_path: Path = target_path / "connections/test_01_conn_stage.csv"

    store.save(
        path=stage_path,
        data={"temp_additional": store.get("conn_local_file")},
        merge=True,
    )
    os.unlink(stage_path)


def test_store(target_path):
    store: Store = Store(target_path)
    store.move(
        path="test_01_conn.yaml",
        dest=target_path / "connections/test_01_conn.yaml",
    )

    stage_path: Path = target_path / "connections/test_01_conn_stage.json"
    stage_path.parent.mkdir(parents=True, exist_ok=True)

    with pytest.raises(TypeError):
        store.save(path=stage_path, data="second", merge=True)

    store.create(path=stage_path)
    assert stage_path.exists()

    store.save(path=stage_path, data=store.get("conn_local_file"))
    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == store.load(path=stage_path)

    assert {} == store.load(
        path=target_path / "connections/test_01_conn_stage_failed.json"
    )
    assert {"foo": "bar"} == store.load(
        path=target_path / "connections/test_01_conn_stage_failed.json",
        default={"foo": "bar"},
    )

    store.save(
        path=stage_path,
        data={"temp_additional": store.get("conn_local_file")},
        merge=True,
    )

    store.delete(
        path=stage_path,
        name="temp_additional",
    )

    assert {
        "alias": "conn_local_file",
        "endpoint": "file:///null/tests/examples/dummy",
        "type": "connection.LocalFileStorage",
    } == store.load(path=stage_path)

    store.delete(
        target_path / "connections/test_01_conn_stage_not_fount.json",
        name="first",
    )
    os.unlink(stage_path)


def test_store_save(target_path):
    store = Store(target_path)
    stage_path: Path = target_path / "connections/test_01_conn_stage_save.json"
    stage_path.parent.mkdir(parents=True, exist_ok=True)

    store.save(
        path=stage_path,
        data={"first": store.get("conn_local_file")} | {"version": 1},
    )
    store.save(
        path=stage_path,
        data={"second": store.get("conn_local_file")} | {"version": 2},
        merge=True,
    )

    assert 2 == store.load(path=stage_path).get("version")

    try:
        store.save(path=stage_path, data="second", merge=True)
    except TypeError:
        assert stage_path.exists()

    assert 2 == store.load(path=stage_path).get("version")
    os.unlink(stage_path)


def test_store_save_list(target_path):
    store = Store(target_path)
    stage_path: Path = target_path / "connections/test_01_conn_stage_list.json"
    stage_path.parent.mkdir(parents=True, exist_ok=True)

    store.save(path=stage_path, data=[{"foo": "bar"}])
    store.save(path=stage_path, data=[{"baz": "bar"}], merge=True)

    assert [{"foo": "bar"}, {"baz": "bar"}] == (store.load(path=stage_path))
    os.unlink(stage_path)


def test_store_save_raise(target_path):
    store: Store = Store(target_path)
    stage_path: Path = target_path / "connections/test_01_conn_stage_raise.json"
    stage_path.parent.mkdir(parents=True, exist_ok=True)
    store.save(
        path=stage_path,
        data={"first": store.get("conn_local_file")} | {"version": 1},
        merge=True,
    )

    with pytest.raises(TypeError):
        store.save(
            path=stage_path,
            data="conn_local_file",
            merge=True,
        )

    store.delete(stage_path, name="first")
    os.unlink(stage_path)


def test_store_json_line(target_path):
    store = StoreToJsonLine(target_path)
    stage_path: Path = target_path / "connections/test_01_conn_stage.line.json"
    stage_path.parent.mkdir(parents=True, exist_ok=True)
    store.save(
        path=stage_path,
        data={"first": store.get("conn_local_file")} | {"version": 1},
        merge=True,
    )
    os.unlink(stage_path)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations


class BaseError(Exception):
    """Base Error Object that use for catch any errors statement of all steps in
    this ``/src`` directory.
    """


class IOBaseError(BaseError):
    """Core Base Error object"""


class ConfigArgumentError(IOBaseError):
    """Error raise for arguments that passing to config object not valid."""


class StoreNotFound(IOBaseError):
    """Error raise for a method not found the config file or data."""


class StoreArgumentError(IOBaseError):
    """Error raise for arguments that passing to store object not valid."""


class RegisterArgumentError(IOBaseError):
    """Error raise for arguments that passing to register object not valid."""



import shutil
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from unittest.mock import patch
from zoneinfo import ZoneInfo

import pytest
import yaml
from ddeutil.io.config import Params
from ddeutil.io.exceptions import RegisterArgumentError
from ddeutil.io.register import Register


@pytest.fixture(scope="module")
def target_path(test_path) -> Iterator[Path]:
    tgt_path: Path = test_path / "register_temp"
    tgt_path.mkdir(exist_ok=True)
    (tgt_path / "conf/demo").mkdir(parents=True, exist_ok=True)
    with open(tgt_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/dummy",
                }
            },
            f,
        )
    yield tgt_path
    shutil.rmtree(tgt_path)


@pytest.fixture(scope="module")
def params(target_path, root_path) -> Params:
    return Params(
        **{
            "paths": {
                "conf": target_path / "conf",
                "data": root_path / "data",
            },
            "stages": {
                "raw": {"format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"},
                "persisted": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
            },
        }
    )


@pytest.fixture(scope="module")
def mock_get_date():
    with patch(
        target="ddeutil.io.register.get_date",
        return_value=datetime(2024, 1, 1, 1, tzinfo=ZoneInfo("UTC")),
    ) as mock:
        yield mock


def test_register(params: Params, mock_get_date):
    assert mock_get_date.mocked

    register = Register(name="demo:conn_local_file", params=params)

    assert str(register) == "(demo:conn_local_file, base)"
    assert repr(register) == "<Register(name='demo:conn_local_file')>"
    assert register.stage == "base"
    assert register.shortname == "clf"
    assert register.fullname == "demo:conn_local_file"

    assert {
        "alias": "conn_local_file",
        "type": "connection.LocalFileStorage",
        "endpoint": "file:///null/tests/examples/dummy",
    } == register.data()

    assert {
        "alias": "62d877a16819c672578d7bded7f5903c",
        "type": "cece9f1b3f4791a04ec3d695cb5ba1a9",
        "endpoint": "0d1db48bb2425db014fc66734508098f",
    } == register.data(hashing=True)

    assert register.changed == 99
    assert str(register.version()) == "0.0.1"

    rsg_raw = register.move(stage="raw")

    assert register.stage != rsg_raw.stage
    assert (
        "62d877a16819c672578d7bded7f5903c"
        == rsg_raw.data(hashing=True)["alias"]
    )

    rsg_persisted = rsg_raw.move(stage="persisted")
    assert rsg_raw.stage != rsg_persisted.stage
    assert (
        "62d877a16819c672578d7bded7f5903c"
        == rsg_persisted.data(hashing=True)["alias"]
    )
    Register.reset(name="demo:conn_local_file", params=params)

    rsg_raw = register.move(stage="raw")
    assert rsg_raw.changed == 99
    assert rsg_raw.timestamp == datetime(2024, 1, 1, 1, tzinfo=ZoneInfo("UTC"))

    register = Register(name="demo:conn_local_file", params=params)
    register.move(stage="raw")

    Register.reset(name="demo:conn_local_file", params=params)


def test_register_compare(params, mock_get_date):
    assert mock_get_date.mocked

    register_01 = Register(name="demo:conn_local_file", params=params)
    register_02 = Register(name="demo:conn_local_file", params=params)
    assert register_01 == register_02
    assert register_01 != "demo:conn_local_file"


def test_register_reset(params: Params):
    Register.reset(name="demo:conn_local_file", params=params)

    register = Register(name="demo:conn_local_file", params=params)
    register.move("raw")
    register.reset(name="demo:conn_local_file", params=params)


def test_register_raise():
    with pytest.raises(RegisterArgumentError):
        Register(name="demo.conn_local_file")

    with pytest.raises(NotImplementedError):
        Register(name="demo:conn_local_file")


def test_register_change_data(params, target_path):
    register = Register(name="demo:conn_local_file", params=params)
    register_raw = register.move(stage="raw", retention=False)
    origin_updt = register_raw.timestamp
    register = Register(name="demo:conn_local_file", params=params)
    assert register.changed == 0

    register_raw = register.move(stage="raw")
    assert origin_updt == register_raw.timestamp

    with open(target_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/new",
                }
            },
            f,
        )
    register = Register(name="demo:conn_local_file", params=params)
    assert register.changed == 1

    register_raw = register.move(stage="raw")
    assert register_raw.data()["__version"] == "v0.0.2"

    with open(target_path / "conf/demo/test_01_conn.yaml", mode="w") as f:
        yaml.dump(
            {
                "conn_local_file": {
                    "type": "connection.LocalFileStorage",
                    "endpoint": "file:///${APP_PATH}/tests/examples/new",
                    "extra": {"foo": "bar"},
                }
            },
            f,
        )

    register = Register(name="demo:conn_local_file", params=params)
    assert register.changed == 2

    register_raw = register.move(stage="raw")
    assert register_raw.data()["__version"] == "v0.1.0"



import shutil
from collections.abc import Iterator
from pathlib import Path

import pytest
from ddeutil.io import touch
from ddeutil.io.files import Fl, compress_lib


@pytest.fixture(scope="module")
def file_mem_path(test_path) -> Iterator[Path]:
    this_path: Path = test_path / "file_memory"
    this_path.mkdir(parents=True, exist_ok=True)

    yield this_path

    shutil.rmtree(this_path)


def test_compress_lib(file_mem_path):
    file = Fl(
        path=file_mem_path / "test_common_mem_file.gz.text",
        compress="gz",
    )

    assert compress_lib("gz").decompress == file.decompress


def test_compress_lib_raise(file_mem_path, encoding):
    file = Fl(
        path=file_mem_path / "test_common_mem_file.foo.text",
        encoding=encoding,
        compress="foo",
    )

    with pytest.raises(NotImplementedError):
        _ = file.decompress


def test_open_file_mem_common(file_mem_path, encoding):
    file = Fl(
        path=file_mem_path / "test_common_mem_file.text",
        encoding=encoding,
    )
    with file.mopen(mode="w") as f:
        f.write("Write data with common file in normal mode on memory")

    with file.mopen(mode="r") as f:
        assert (
            b"Write data with common file in normal mode on memory" == f.read()
        )


def test_open_file_mem_common_raise(file_mem_path, encoding):
    file = Fl(
        path=file_mem_path / "test_common_mem_file_raise.text",
        encoding=encoding,
    )
    with pytest.raises(FileNotFoundError):
        with file.mopen(mode="r") as f:
            f.read()

    touch(file_mem_path / "test_common_mem_file_raise.text")
    with file.mopen(mode="r") as f:
        assert f.read() == ""


def test_open_file_mem_common_gzip(file_mem_path, encoding):
    file = Fl(
        path=file_mem_path / "test_common_mem_file.gz.text",
        encoding=encoding,
        compress="gzip",
    )
    with file.mopen(mode="w") as f:
        f.write("Write data with common file in gzip mode on memory")

    with file.mopen(mode="r") as f:
        rs = compress_lib("gzip").decompress(f.read())

    assert b"Write data with common file in gzip mode on memory" == rs


def test_open_file_mem_common_xz(file_mem_path, encoding):
    file = Fl(
        path=file_mem_path / "test_common_mem_file.xz.text",
        encoding=encoding,
        compress="xz",
    )
    with file.mopen(mode="w") as f:
        f.write("Write data with common file in xz mode on memory")

    with file.mopen(mode="r") as f:
        rs = compress_lib("xz").decompress(f.read())

    assert b"Write data with common file in xz mode on memory" == rs


def test_open_file_mem_common_bz2(file_mem_path, encoding):
    file = Fl(
        path=file_mem_path / "test_common_mem_file.bz2.text",
        encoding=encoding,
        compress="bz2",
    )
    with file.mopen(mode="w") as f:
        f.write("Write data with common file in bz2 mode on memory")

    with file.mopen(mode="r") as f:
        rs = compress_lib("bz2").decompress(f.read())

    assert b"Write data with common file in bz2 mode on memory" == rs



import gzip
import json
import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from textwrap import dedent

import pytest
from ddeutil.io.files import JsonEnvFl, JsonFl, JsonLineFl


@pytest.fixture(scope="module")
def json_path(test_path) -> Iterator[Path]:
    this_path: Path = test_path / "json"
    this_path.mkdir(parents=True, exist_ok=True)

    with open(this_path / "test_simple.json", mode="w", encoding="utf-8") as f:
        f.write(
            dedent(
                """
            {
                "config": {
                    // Comment this line ...
                    "value": "foo"
                }
            }
            """
            ).strip()
        )

    with open(
        this_path / "test_simple_raise.json", mode="w", encoding="utf-8"
    ) as f:
        f.write(
            dedent(
                """
            {
                "config": {
                    // Comment this line ...
                    "value": "foo",
                    "value": "${TEST_JSON_ENV}"
                }
            """
            ).strip()
        )

    with open(this_path / "test_env.json", mode="w", encoding="utf-8") as f:
        f.write(
            dedent(
                """
            {
                "config": {
                    // Comment this line ...
                    "value": "foo is ${TEST_JSON_ENV}"
                }
            }
            """
            ).strip()
        )

    yield this_path

    shutil.rmtree(this_path)


def test_files_open_json(json_path):
    assert {"config": {"value": "foo"}} == JsonFl(
        path=json_path / "test_simple.json"
    ).read()


def test_files_open_json_raise(json_path):
    with pytest.raises(json.decoder.JSONDecodeError):
        JsonFl(path=json_path / "test_simple_raise.json").read()


def test_files_open_json_write(json_path):
    JsonFl(path=json_path / "test_simple_write.json").write(
        {"config": {"value": "foo"}},
        indent=0,
    )
    with open(json_path / "test_simple_write.json") as f:
        assert f.read() == '{\n"config": {\n"value": "foo"\n}\n}'


def test_files_open_json_write_compress(json_path):
    JsonFl(
        path=json_path / "test_simple_write.gz.json",
        compress="gz",
    ).write(
        {"config": {"value": "foo"}},
        indent=0,
    )
    with gzip.open(json_path / "test_simple_write.gz.json") as f:
        assert f.read() == b'{"config": {"value": "foo"}}'


def test_files_open_json_env_read(json_path):
    os.environ["TEST_JSON_ENV"] = "FOO"
    assert {"config": {"value": "foo is FOO"}} == JsonEnvFl(
        path=json_path / "test_env.json"
    ).read()


def test_files_open_json_env_read_raise(json_path):
    with pytest.raises(json.decoder.JSONDecodeError):
        JsonEnvFl(path=json_path / "test_simple_raise.json").read()


def test_files_open_json_line(json_path):
    with open(json_path / "test_write.line.json", mode="w") as f:
        f.write("")

    assert [] == JsonLineFl(path=json_path / "test_write.line.json").read()

    JsonLineFl(path=json_path / "test_write.line.json").write({"line": 1})
    JsonLineFl(path=json_path / "test_write.line.json").write(
        [{"line": 2}, {"line": 3}], mode="a"
    )

    assert [{"line": 1}, {"line": 2}, {"line": 3}] == JsonLineFl(
        path=json_path / "test_write.line.json"
    ).read()


def test_files_open_json_line_raise(json_path):
    with open(json_path / "test_write_raise.line.json", mode="w") as f:
        f.write("\n")

    with pytest.raises(json.decoder.JSONDecodeError):
        JsonLineFl(path=json_path / "test_write_raise.line.json").read()

    with pytest.raises(ValueError):
        JsonLineFl(path=json_path / "test_write_raise.line.json").write([])



import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from textwrap import dedent

import pytest
from ddeutil.io.files import TomlEnvFl, TomlFl


@pytest.fixture(scope="module")
def toml_path(test_path) -> Iterator[Path]:
    this_path: Path = test_path / "toml"
    this_path.mkdir(parents=True, exist_ok=True)

    with open(this_path / "test_simple.toml", mode="w", encoding="utf-8") as f:
        f.write(
            dedent(
                """[config]
            # Comment this line ...
            value = "foo"
            """
            ).strip()
        )

    with open(this_path / "test_env.toml", mode="w", encoding="utf-8") as f:
        f.write(
            dedent(
                """[config]
            # Comment this line ...
            value = "foo is ${TEST_TOML_ENV}"
            """
            ).strip()
        )

    yield this_path

    shutil.rmtree(this_path)


def test_files_open_toml_read(toml_path):
    assert {"config": {"value": "foo"}} == TomlFl(
        path=toml_path / "test_simple.toml"
    ).read()


def test_files_open_toml_write(toml_path):
    TomlFl(path=toml_path / "test_simple_write.toml").write(
        {"config": {"value": "foo"}}
    )
    with open(toml_path / "test_simple_write.toml") as f:
        assert f.read() == ("[config]\n" 'value = "foo"\n')


def test_files_open_toml_env_read(toml_path):
    os.environ["TEST_TOML_ENV"] = "FOO"
    assert {"config": {"value": "foo is FOO"}} == TomlEnvFl(
        path=toml_path / "test_env.toml"
    ).read()



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
"""Store objects that use to keep context configuration data with different
versions. This module will provide standard and abstraction objects for your
customize usage.

    *   Store           : Store with Yaml open file and use stage with Json.
    *   StoreJsonToCsv  : Store with Json open file and use stage with Yaml.

    Store will keep data with 2 stages, that mean data have data layer and stage
layer.
"""
from __future__ import annotations

import abc
import inspect
import logging
import shutil
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import Any, ClassVar, Optional, Union

from .__type import AnyData, TupleStr
from .config import VERSION_DEFAULT
from .files import (
    CsvPipeFl,
    Fl,
    JsonEnvFl,
    JsonFl,
    JsonLineFl,
    YamlEnvFl,
)
from .paths import PathSearch
from .utils import rm

__all__: TupleStr = (
    "BaseStore",
    "Store",
    "StoreJsonToCsv",
    "StoreToJsonLine",
)


class BaseStore(abc.ABC):
    """Base Store File object for getting data with `.yaml` format (default
    format for a config file) and mapping environment variables to the content
    data.

        This object implement only source file without stage open file.

        Base Store Adapter abstract class for any config sub-class that should
    implement necessary methods for unity usage and dynamic config backend
    changing scenario.

    :param path:
    :param compress:
    """

    open_file: ClassVar[type[Fl]] = YamlEnvFl
    included_file_fmt: ClassVar[TupleStr] = ("*.yml", "*.yaml")
    excluded_file_fmt: ClassVar[TupleStr] = ("*.json", "*.toml", "*.csv")

    def __init__(
        self,
        path: Union[str, Path],
        *,
        compress: Optional[str] = None,
    ) -> None:
        self.path: Path = Path(path) if isinstance(path, str) else path
        self.compress: Optional[str] = compress

        # NOTE: Create parent dir and skip if it already exists
        if not self.path.exists():
            self.path.mkdir(parents=True)

    def get(self, name: str, *, order: int = 1) -> AnyData:
        """Return configuration data from name of the config that already adding
        `alias` key with this input name.

        :param name: A name of config key that want to search in the path.
        :type name: str
        :param order: An order number that want to get from ordered list
            of duplicate data.
        :type order: int (Default=1)

        :rtype: AnyData
        :returns: The loaded context data from the open file read method.
        """
        rs: list[dict[Any, Any]]
        if not (
            rs := [
                {"alias": name} | data
                for file in self.ls(excluded=self.excluded_file_fmt)
                if (
                    data := (
                        self.open_file(path=file, compress=self.compress)
                        .read()
                        .get(name)
                    )
                )
            ]
        ):
            return {}

        try:
            if order > len(rs):
                raise IndexError(
                    "Order argument should be less or equal than len of "
                    "data that exist in the store path."
                )
            return sorted(
                rs,
                key=lambda x: (
                    datetime.fromisoformat(x.get("version", VERSION_DEFAULT)),
                    len(x),
                ),
                reverse=False,
            )[-order]
        except IndexError:
            logging.warning(
                f"Does not get config data {name!r} with passing order: "
                f"-{order}"
            )
            return {}

    def ls(
        self,
        path: Optional[str] = None,
        name: Optional[str] = None,
        *,
        excluded: Optional[Union[list[str], tuple[str, ...]]] = None,
    ) -> Iterator[Path]:
        """Return all files that already exist in the store path.

        :param path: A specific root path that want to list.
        :param name: A filename pattern that want to list.
        :param excluded: A list of excluded filenames.
        :rtype: Iterator[Path]
        """
        yield from filter(
            lambda x: x.is_file(),
            (
                PathSearch(
                    root=(path or self.path),
                    exclude=excluded,
                ).pick(filename=(name or "*"))
            ),
        )

    def move(self, path: Union[str, Path], dest: Path) -> None:
        """Copy filename inside this config path to the destination path.

        :param path: A child path that exists in this store path.
        :param dest: A destination path.
        """
        if not dest.parent.exists():
            dest.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(self.path / path, dest)

    @abc.abstractmethod
    def load(self, name: str) -> dict[str, Any]:  # pragma: no cov
        raise NotImplementedError()

    @abc.abstractmethod
    def save(
        self, name: str, data: dict, *, merge: bool = False
    ) -> None:  # pragma: no cov
        raise NotImplementedError()

    @abc.abstractmethod
    def delete(self, path: str, name: str) -> None:  # pragma: no cov
        raise NotImplementedError()

    @abc.abstractmethod
    def create(self, name: str, **kwargs) -> None:  # pragma: no cov
        raise NotImplementedError()


class Store(BaseStore):
    """Store File Loading Object for get data from configuration and stage.

    :param path: A path of files to action.
    :param compress: A compress type of action file.
    """

    open_file: ClassVar[type[Fl]] = YamlEnvFl
    open_file_stg: ClassVar[type[Fl]] = JsonFl

    def __init__(
        self,
        path: Union[str, Path],
        *,
        compress: Optional[str] = None,
    ) -> None:
        """Main initialize of config file loading object."""
        super().__init__(path, compress=compress)

    def load(
        self, path: Union[str, Path], *, default: AnyData = None
    ) -> AnyData:
        """Return content data from file with filename, default empty dict.

        :rtype: AnyData
        """
        try:
            return self.open_file_stg(path=path, compress=self.compress).read()
        except FileNotFoundError:
            return default if (default is not None) else {}

    def save(
        self,
        path: Union[str, Path],
        data: AnyData,
        *,
        merge: bool = False,
    ) -> None:
        """Write content data to file with filename. If merge is true, it will
        load the current data from saving file and merge the incoming data
        together before re-write the file.

        :param path:
        :param data:
        :param merge:
        """
        if not merge:
            self.open_file_stg(path, compress=self.compress).write(data)
            logging.debug(f"Start writing data to {path}")
            return
        elif merge and (
            "mode"
            in inspect.getfullargspec(self.open_file_stg.write).annotations
        ):
            self.open_file_stg(path, compress=self.compress).write(
                **{"data": data, "mode": "a"}
            )
            return

        all_data: AnyData = self.load(path=path)
        try:
            if isinstance(all_data, list):
                rs: list[AnyData] = all_data
                (rs.append(data) if isinstance(data, dict) else rs.extend(data))
            else:
                rs: dict = all_data | data

            # NOTE: Writing data to the stage layer
            self.open_file_stg(path, compress=self.compress).write(rs)
        except TypeError:
            # NOTE: Remove the previous saving file path for rollback.
            rm(path=path, force_raise=False)
            if all_data:
                self.open_file_stg(path, compress=self.compress).write(
                    all_data,
                )
            raise

    def delete(self, path: Union[str, Path], name: str) -> None:
        """Remove data by name insided the staging file with filename.

        :param path:
        :param name:
        """
        # NOTE: Remove data with the input name key if it exists.
        if all_data := self.load(path=path):
            all_data.pop(name, None)
            (self.open_file_stg(path, compress=self.compress).write(all_data))

    def create(
        self, path: Union[str, Path], *, initial_data: AnyData = None
    ) -> None:
        """Create file with an input filename to the store path. This method
        allow to create with initial data.

        :param path:
        :param initial_data:
        :type initial_data: AnyData
        :rtype: None
        """
        if not path.exists():
            self.save(
                path=path,
                data=(initial_data or {}),
                merge=False,
            )


class StoreJsonToCsv(Store):
    """Store object that getting the Json context data and save it to stage with
    CSV file format.
    """

    open_file: ClassVar[type[Fl]] = JsonEnvFl
    open_file_stg: ClassVar[type[Fl]] = CsvPipeFl
    included_file_fmt: ClassVar[TupleStr] = ("*.json",)
    excluded_file_fmt: ClassVar[TupleStr] = (
        "*.yml",
        "*.yaml",
        "*.toml",
        "*.csv",
    )


class StoreToJsonLine(Store):
    """Store object that getting the YAML context data and save it to stage with
    Json line file format.
    """

    open_file_stg: ClassVar[type[Fl]] = JsonLineFl



import shutil
from collections.abc import Generator
from pathlib import Path

import pytest
from ddeutil.io.paths import PathSearch, ls, replace_sep
from ddeutil.io.utils import touch


@pytest.fixture(scope="module")
def make_empty_path(test_path: Path) -> Generator[Path, None, None]:
    path_search = test_path / "test_empty_path_search"
    path_search.mkdir(exist_ok=True)

    yield path_search

    shutil.rmtree(path_search)


@pytest.fixture(scope="module")
def make_path(test_path: Path) -> Generator[Path, None, None]:
    path_search: Path = test_path / "test_path_search"
    path_search.mkdir(exist_ok=True)

    touch(path_search / "00_01_test.text")
    (path_search / "dir01").mkdir(exist_ok=True)
    touch(path_search / "dir01" / "01_01_test.text")
    touch(path_search / "dir01" / "01_02_test.text")
    (path_search / "dir02").mkdir(exist_ok=True)
    touch(path_search / "dir02" / "02_01_test.text")

    yield path_search

    shutil.rmtree(path_search)


def test_base_path_search_empty(make_empty_path):
    ps = PathSearch(make_empty_path)
    assert [] == ps.files
    assert 1 == ps.level


def test_base_path_search_raise(make_empty_path):
    with pytest.raises(FileNotFoundError):
        PathSearch(make_empty_path / "demo")


def test_base_path_search(make_path):
    ps = PathSearch(make_path)
    assert {
        make_path / "00_01_test.text",
        make_path / "dir01/01_01_test.text",
        make_path / "dir01/01_02_test.text",
        make_path / "dir02/02_01_test.text",
    } == set(ps.files)

    ps = PathSearch(make_path, exclude=["dir02"])
    assert {
        make_path / "00_01_test.text",
        make_path / "dir01/01_01_test.text",
        make_path / "dir01/01_02_test.text",
    } == set(ps.files)


@pytest.fixture(scope="module")
def make_ls(test_path: Path) -> Generator[Path, None, None]:
    path_search: Path = test_path / "test_path_ls"
    path_search.mkdir(exist_ok=True)

    touch(path_search / "00_01_test.yml")
    (path_search / "dir01").mkdir(exist_ok=True)
    touch(path_search / "dir01" / "01_01_test.yml")
    touch(path_search / "dir01" / "01_02_test.yml")

    (path_search / "dir02").mkdir(exist_ok=True)
    touch(path_search / "dir02" / "02_01_test.yml")
    touch(path_search / "dir02" / "02_01_test.json")
    touch(path_search / "dir02" / "02_02_test_ignore.yml")
    touch(path_search / "dir02" / "02_03_test.yml")

    (path_search / "dir02/tests").mkdir(exist_ok=True)
    touch(path_search / "dir02/tests" / "02_01_01_demo.yml")

    (path_search / "tests").mkdir(exist_ok=True)
    touch(path_search / "tests" / "03_01_test.yml")
    touch(path_search / "tests" / "03_02_test.yml")

    with open(path_search / ".ignore", mode="w") as f:
        f.write("tests\n")
        f.write("*.json\n")
        f.write("*_ignore.yml\n")
        f.write("02_03_*\n")

    yield path_search

    shutil.rmtree(path_search)


def test_ls(make_ls):
    files = ls(make_ls, ignore_file=".ignore")

    assert [replace_sep(str(f.relative_to(make_ls))) for f in files] == [
        "00_01_test.yml",
        "dir01/01_01_test.yml",
        "dir01/01_02_test.yml",
        "dir02/02_01_test.yml",
    ]


def test_ls_empty(test_path):
    path_search: Path = test_path / "test_path_ls_empty"
    path_search.mkdir(exist_ok=True)

    files = ls(path_search, ignore_file=".ignore")
    assert list(files) == []

    shutil.rmtree(path_search)



import shutil
from collections.abc import Iterator
from pathlib import Path

import pytest
from ddeutil.io.files import CsvFl, CsvPipeFl


@pytest.fixture(scope="module")
def csv_path(test_path) -> Iterator[Path]:
    this_path: Path = test_path / "csv"
    this_path.mkdir(parents=True, exist_ok=True)

    yield this_path

    shutil.rmtree(this_path)


@pytest.fixture(scope="module")
def csv_data() -> list[dict[str, str]]:
    return [
        {"Col01": "A", "Col02": "1", "Col03": "test1"},
        {"Col01": "B", "Col02": "2", "Col03": "test2"},
        {"Col01": "C", "Col02": "3", "Col03": "test3"},
    ]


def test_files_open_csv(csv_path, csv_data):
    CsvFl(csv_path / "test_file.csv").write(csv_data)
    assert csv_data == CsvFl(csv_path / "test_file.csv").read()


def test_files_open_csv_dict(csv_path, csv_data):
    CsvFl(csv_path / "test_file.dict.csv").write(
        {"Col01": "A", "Col02": "1", "Col03": "test1"}
    )
    assert [{"Col01": "A", "Col02": "1", "Col03": "test1"}] == CsvFl(
        csv_path / "test_file.dict.csv"
    ).read()

    CsvFl(csv_path / "test_file.dict.csv").write(
        {"Col01": "B", "Col02": "2", "Col03": "test2"}, mode="a"
    )
    CsvFl(csv_path / "test_file.dict.csv").write(
        [{"Col01": "C", "Col02": "3", "Col03": "test3"}], mode="a"
    )

    with pytest.raises(ValueError):
        CsvFl(csv_path / "test_file.dict.csv").write([])

    assert csv_data == CsvFl(csv_path / "test_file.dict.csv").read()


def test_files_open_csv_pipe_dict(csv_path, csv_data):
    CsvPipeFl(csv_path / "test_file_pipe.dict.csv").write(
        {"Col01": "A", "Col02": "1", "Col03": "test1"}
    )
    assert [{"Col01": "A", "Col02": "1", "Col03": "test1"}] == CsvPipeFl(
        csv_path / "test_file_pipe.dict.csv"
    ).read()

    CsvPipeFl(csv_path / "test_file_pipe.dict.csv").write(
        {"Col01": "B", "Col02": "2", "Col03": "test2"}, mode="a"
    )
    CsvPipeFl(csv_path / "test_file_pipe.dict.csv").write(
        [{"Col01": "C", "Col02": "3", "Col03": "test3"}], mode="a"
    )

    with pytest.raises(ValueError):
        CsvPipeFl(csv_path / "test_file_pipe.dict.csv").write([])

    assert csv_data == CsvPipeFl(csv_path / "test_file_pipe.dict.csv").read()


def test_files_open_csv_not_support(csv_path):
    with open(csv_path / "test_file_raise.csv", mode="w") as f:
        f.write("...")
    assert [] == CsvFl(csv_path / "test_file_raise.csv").read()



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from typing import IO, AnyStr, Callable, Optional, TypeVar, Union

from ddeutil.core import convert, import_string

from .__regex import RegexConf

T = TypeVar("T")


def rm(
    path: Union[str, Path],
    is_dir: bool = False,
    force_raise: bool = True,
) -> None:  # pragma: no cov
    """Remove a file or dir from an input path.

    :param path: A path of file or dir that want to remove.
    :param is_dir: A flag that tell this input path is dir or not.
    :param force_raise: A flag that disable raise error if it not remove.
    """
    path: Path = Path(path) if isinstance(path, str) else path
    if os.path.isfile(path) or os.path.islink(path):
        os.remove(path)
    elif os.path.isdir(path) and is_dir:
        shutil.rmtree(path)
    else:
        if force_raise:
            raise ValueError(
                f"Path {path!r} is not a file{' or dir' if is_dir else ''}."
            )


def touch(path: Union[str, Path], times=None) -> None:  # pragma: no cov
    """Create an empty file with specific name and modified time of path it an
    input times was set.

    :param path: A file path that want to create.
    :param times: A time that want to adjust modified time.
    """
    file_handle = open(path, mode="a")
    try:
        os.utime(path, times)
    finally:
        file_handle.close()


def template_secret(value: T, secrets: dict[str, str]) -> T:
    """Map the secret value to an any input data.

    :param value: An input value that want to map secrets
    :param secrets: A mapping of value secrets that use to replace.
    :type secrets: dict[str, str]

    Examples:
        >>> template_secret("s3://@secrets{foo}", secrets={"foo": "bar"})
        's3://bar'
    """
    if isinstance(value, dict):
        return {k: template_secret(value[k], secrets) for k in value}
    elif isinstance(value, (list, tuple)):
        return type(value)([template_secret(i, secrets) for i in value])
    elif not isinstance(value, str):
        return value
    for search in RegexConf.RE_SECRETS.finditer(value):
        searches: dict = search.groupdict()
        if "." in (br := searches["braced"]):
            raise ValueError(
                f"The @secrets: {br!r}, should not contain dot ('.') char"
            )
        value: str = value.replace(
            searches["search"],
            secrets.get(br.strip(), searches["braced_default"]),
        )
    return value


def template_func(value: T) -> T:
    """Map the function result to configuration data.

    :param value: A data that want to map imported function with arguments.

    Examples:
        >>> template_func(
        ...     "Test @function{ddeutil.io.files.add_newline:'a',newline='|'}"
        ... )
        'Test a|'
    """
    if isinstance(value, dict):
        return {k: template_func(value[k]) for k in value}
    elif isinstance(value, (list, tuple)):
        return type(value)([template_func(i) for i in value])
    elif not isinstance(value, str):
        return value
    for search in RegexConf.RE_FUNCTION.finditer(value):
        searches: dict = search.groupdict()
        if not callable(_fn := import_string(searches["function"])):
            raise ValueError(
                f'The @function: {searches["function"]!r} is not callable.',
            )
        args, kwargs = convert.str2args(searches["arguments"])
        value: str = value.replace(searches["search"], _fn(*args, **kwargs))
    return value


def map_func(value: T, func: Callable[[str], str]) -> T:
    """Map any function from input argument to configuration data.

    Examples:
        >>> map_func({"foo": "bar"}, lambda x: x + "!")
        {'foo': 'bar!'}
        >>> map_func(("foo", "bar"), lambda x: x + "!")
        ('foo!', 'bar!')
    """
    if isinstance(value, dict):
        return {k: map_func(value[k], func) for k in value}
    elif isinstance(value, (list, tuple)):
        return type(value)([map_func(i, func) for i in value])
    elif not isinstance(value, str):
        return value
    return func(value)


def add_newline(text: str, newline: Optional[str] = None) -> str:
    """Add newline to a text value.

    :param text: A text value that want to add newline.
    :param newline: A newline value that want to use.

    :rtype: str
    :returns: A newline added text.
    """
    nl: str = newline or "\n"
    return f"{text}{nl}" if not text.endswith(nl) else text


def search_env_replace(
    contents: str,
    *,
    raise_if_default_not_exists: bool = False,
    default: str = "null",
    escape: str = "ESC",
    caller: Callable[[str], str] = (lambda x: x),
) -> str:
    """Prepare content data before parse to any file parsing object.

    :param contents: A string content that want to format with env vars
    :type contents: str
    :param raise_if_default_not_exists: A flag that will allow this function
        raise the error when default of env var does not set from contents.
    :type raise_if_default_not_exists: bool(=False)
    :param default: a default value.
    :type default: str(='null')
    :param escape: a escape value that use for initial replace when found escape
        char on searching.
    :type escape: str(='ESC')
    :param caller: a prepare function that will execute before replace env var.
    :type caller: Callable[[str], str]

    :rtype: str
    :returns: A prepared content data.

    Examples:

        >>> import os
        >>> os.environ["NAME"] = 'foo'
        >>> search_env_replace("Hello ${NAME}")
        'Hello foo'
    """
    shifting: int = 0
    replaces: dict = {}
    replaces_esc: dict = {}
    for content in RegexConf.RE_ENV_SEARCH.finditer(contents):
        search: str = content.group(1)
        if not (_escaped := content.group("escaped")):
            var: str = content.group("braced")
            _braced_default: str = content.group("braced_default")
            if not _braced_default and raise_if_default_not_exists:
                raise ValueError(
                    f"Could not find default value for {var} in the contents"
                )
            elif not var:
                raise ValueError(
                    f"Value {search!r} in the contents file has something "
                    f"wrong with regular expression"
                )
            replaces[search] = caller(
                os.environ.get(var, _braced_default) or default
            )
        elif "$" in _escaped:
            span = content.span()
            search = f"${{{escape}{_escaped}}}"
            contents = (
                contents[: (span[0] + shifting)]
                + search
                + contents[(span[1] + shifting) :]
            )
            shifting += len(search) - (span[1] - span[0])
            replaces_esc[search] = "$"
    for _replace in sorted(replaces, reverse=True):
        contents = contents.replace(_replace, replaces[_replace])
    for _replace in sorted(replaces_esc, reverse=True):
        contents = contents.replace(_replace, replaces_esc[_replace])
    return contents


def search_env(
    contents: str,
    *,
    keep_newline: bool = False,
    default: Optional[str] = None,
) -> dict[str, str]:
    """Prepare content data from `.env` file before load to the OS environment
    variables.

    :param contents: A string content in the `.env` file
    :type contents: str
    :param keep_newline: A flag that filter out a newline
    :type keep_newline: bool(=False)
    :param default: A default value that use if it does not exists
    :type default: str | None(=None)

    :rtype: dict[str, str]
    :returns: A mapping of name and value of env variable

    Note:
        This function reference code from python-dotenv package. I will use this
    instead install this package. Because I want to enhance serialize step that
    fit with my package. (https://github.com/theskumar/python-dotenv)

    Examples:
        >>> search_env("Data='demo'\\nfoo=bar")
        {'Data': 'demo', 'foo': 'bar'}
        >>> search_env("Data='demo'\\n# foo=bar\\nhello=${Data}-2")
        {'Data': 'demo', 'hello': 'demo-2'}
    """
    _default: str = default or ""
    env: dict[str, str] = {}
    for content in RegexConf.RE_DOTENV.finditer(contents):
        name: str = content.group("name")

        # NOTE: Remove leading/trailing whitespace
        _value: str = (content.group("value") or "").strip()

        if not _value or _value in ("''", '""'):
            raise ValueError(
                f"Value {name!r} in `.env` file does not set value "
                f"of variable"
            )
        value: str = _value if keep_newline else "".join(_value.splitlines())
        quoted: Optional[str] = None

        # NOTE: Remove surrounding quotes
        if m2 := RegexConf.RE_ENV_VALUE_QUOTED.match(value):
            quoted: str = m2.group("quoted")
            value: str = m2.group("value")

        if quoted == "'":
            env[name] = value
            continue
        elif quoted == '"':
            # NOTE: Unescape all chars except $ so variables
            #   can be escaped properly
            value: str = RegexConf.RE_ENV_ESCAPE.sub(r"\1", value)

        # NOTE: Substitute variables in a value
        env[name] = __search_var(value, env, default=_default)
    return env


def __search_var(
    value: str,
    env: dict[str, str],
    *,
    default: Optional[str] = None,
) -> str:
    """Search variable on the string content.

    :param value: a string value that want to search env variable.
    :type value: str
    :param env: a pair of env values that keep in memory dict.
    :type env: dict[str, str]
    :param default: a default value if it does not found on env vars.
    :type default: str | None(=None)

    :rtype: str
    :returns: A searched value from env veriables.

    Examples:
        >>> __search_var("Test ${VAR}", {"VAR": "foo"})
        'Test foo'
        >>> __search_var("Test ${VAR2}", {"VAR": "foo"})
        'Test '
        >>> __search_var("Test ${VAR2}", {"VAR": "foo"}, default="bar")
        'Test bar'
        >>> import os
        >>> os.environ["VAR2"] = "baz"
        >>> __search_var("Test ${VAR2}", {"VAR": "foo"}, default="bar")
        'Test baz'
    """
    _default: str = default or ""
    for sub_content in RegexConf.RE_DOTENV_VAR.findall(value):
        replace: str = "".join(sub_content[1:-1])
        if sub_content[0] != "\\":
            # NOTE: Replace it with the value from the environment
            replace: str = env.get(
                sub_content[-1],
                os.environ.get(sub_content[-1], _default),
            )
        value: str = value.replace("".join(sub_content[:-1]), replace)
    return value


def reverse_readline(
    f: IO,
    buf_size: int = 8192,
) -> Iterator[AnyStr]:  # pragma: no cov
    """A generator that returns the lines of a file in reverse order

    Reference:
        - https://stackoverflow.com/questions/2301789/ -
            how-to-read-a-file-in-reverse-order
        - https://stackoverflow.com/a/23646049/8776239
    """
    segment: AnyStr | None = None
    offset: int = 0
    f.seek(0, os.SEEK_END)
    file_size = remaining_size = f.tell()

    while remaining_size > 0:
        offset = min(file_size, offset + buf_size)
        f.seek(file_size - offset)
        buffer: AnyStr = f.read(min(remaining_size, buf_size))
        remaining_size -= buf_size
        lines: list[AnyStr] = buffer.splitlines(True)

        # NOTE: the first line of the buffer is probably not a complete line so
        #   we'll save it and append it to the last line of the next buffer
        #   we read
        if segment is not None:

            # NOTE: if the previous chunk starts right from the beginning of
            #   line do not concat the segment to the last line of new chunk
            #   instead, yield the segment first
            if buffer[-1] == "\n":
                yield segment
            else:
                lines[-1] += segment
        segment: AnyStr = lines[0]
        for index in range(len(lines) - 1, 0, -1):
            if len(lines[index]):
                yield lines[index]

    # WARNING: Don't yield None if the file was empty
    if segment is not None:
        yield segment



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import os
import re
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    from typing_extensions import Self

from .__type import TupleStr
from .exceptions import ConfigArgumentError
from .files import YamlEnvFl

UPDATE_KEY: str = "__updt"
VERSION_KEY: str = "__version"
VERSION_DEFAULT: str = "1990-01-01"
DATE_FMT: str = "%Y-%m-%d %H:%M:%S"
DATE_LOG_FMT: str = "%Y%m%d%H%M%S"
FMT_NAMES: TupleStr = (
    "naming",
    "domain",
    "environ",
    "timestamp",
    "version",
    "compress",
    "extension",
)
RULE_NECESSARY_KEYS: TupleStr = (
    "timestamp",
    "compress",
)


def get_root_path() -> Path:
    """Return the root path for this package that receive env var with
    ``IO_REGISTER_ROOT_PATH`` key.

    :rtype: Path
    """
    if (root := os.getenv("IO_REGISTER_ROOT_PATH")) is not None:
        return Path(root)
    return Path()


@dataclass(frozen=True)
class Rule:
    """Rule dataclass that keep rule setting data for Register object.

    Examples:
        >>> rule = {
        ...     "timestamp": {"minutes": 15},
        ...     "excluded": [],
        ...     "compress": None,
        ... }
    """

    timestamp: dict[str, int] = field(default_factory=dict)
    excluded: list = field(default_factory=list)
    compress: Optional[str] = field(default=None)


@dataclass
class Stage:
    """Stage dataclass that keep stage data for transition data.

    Examples:
        >>> stage = {
        ...     "raw": {
        ...         "format": "",
        ...         "rule": { ... },
        ...     },
        ... }
    """

    format: str
    alias: Optional[str] = field(default=None)
    rule: Rule = field(default_factory=Rule)
    layer: int = field(default=0)

    def __post_init__(self) -> Self:
        """Post initialize method of this stage dataclass."""
        # NOTE: Nested Rule object on the rule field.
        if isinstance(self.rule, dict):
            self.rule: Rule = Rule(**self.rule)

        # VALIDATE: Check the name in format string should contain any format
        #   name.
        if not (
            _searches := re.findall(
                r"{\s?(?P<name>\w+):?(?P<format>[^{}]+)?\s?}",
                self.format,
            )
        ):
            raise ConfigArgumentError(
                f"This `{self.alias}` stage format dose not include any format "
                f"name, the stage file was duplicated."
            )

        # VALIDATE: Check the name in format string should exist in `FMT_NAMES`.
        if any((_search[0] not in FMT_NAMES) for _search in _searches):
            raise ConfigArgumentError(
                "The stage has an unsupported format name.",
            )

        # VALIDATE: Validate a format of stage that relate with rules.
        for rule_key in RULE_NECESSARY_KEYS:
            if getattr(self.rule, rule_key, None) and (
                not re.search(rf"{rule_key}", self.format)
            ):
                raise ConfigArgumentError(
                    f"This stage rule was set `{rule_key}` property but does "
                    f"not have a `{rule_key}` format name in the format."
                )
        return self


base_stage: Stage = Stage(
    **{
        "format": "{naming}_{timestamp}",
        "alias": "base",
    }
)


@dataclass
class Paths:
    """Paths dataclass that keep necessary paths for register or loading object.

    Examples:
        >>> path = {
        ...     "root": "./",
        ...     "data": "./data",
        ...     "conf": "./config",
        ... }
    """

    root: Path = field(default_factory=get_root_path)
    data: Path = field(default=None)
    conf: Path = field(default=None)

    def __post_init__(self) -> Self:
        """Post initialize for prepare paths that receive with string value."""
        if isinstance(self.root, str):
            self.root = Path(self.root)

        if self.data is None:
            self.data = self.root / "data"
        elif isinstance(self.data, str):
            self.data = Path(self.data)

        if self.conf is None:
            self.conf = self.root / "conf"
        elif isinstance(self.conf, str):
            self.conf = Path(self.conf)

        return self


@dataclass
class Params:
    """Params dataclass."""

    stages: dict[str, Stage]
    paths: Paths = field(default_factory=Paths)

    @classmethod
    def from_toml(cls, path: Union[Path]) -> Self:
        """Read params from .toml file"""
        import toml

        with open(
            path or "./io-register.toml",
            encoding="utf-8",
        ) as f:
            data: dict[str, Any] = (
                toml.load(f).get("tool", {}).get("io", {}).get("register", {})
            )
        return Params(**data)

    @classmethod
    def from_yaml(cls, path: Union[Path]) -> Self:
        """Read params from .yaml file"""
        return cls(
            **YamlEnvFl(path or "./io-register.yaml")
            .read()
            .get("tool", {})
            .get("io", {})
            .get("register", {})
        )

    def __post_init__(self) -> Self:
        """Post initialize for prepare params."""
        for index, k in enumerate(self.stages, start=1):
            stage: Union[dict[str, Any], Stage] = self.stages[k]
            if isinstance(stage, dict):
                layer: int = stage.get("layer") or index
                self.stages[k] = Stage(**(stage | {"alias": k, "layer": layer}))
            elif isinstance(stage, Stage):
                layer: int = stage.layer or index
                stage.alias = k
                stage.layer = layer
                self.stages[k] = stage
            else:
                raise ConfigArgumentError(f"The stage: {k} does not valid type")

        if isinstance(self.paths, dict):
            self.paths: Paths = Paths(**self.paths)
        return self

    @property
    def stage_final(self) -> str:
        """Return the final stage name that ordered from layer value."""
        return max(self.stages.items(), key=lambda i: i[1].layer)[0]

    @property
    def stage_first(self) -> str:
        """Return the first stage name that ordered from layer value which
        does not be the base stage.
        """
        return min(self.stages.items(), key=lambda i: i[1].layer)[0]

    def get_stage(self, name: str) -> Stage:
        """Return Stage model that match with stage name. If an input stage
        value equal 'base', it will return the default stage model.

        :param name: A stage name that want to get from this params.
        :type name: str
        """
        if name == "base":
            return base_stage
        elif name not in self.stages:
            raise ConfigArgumentError(
                f"Cannot get stage: {name!r} cause it does not exists",
            )
        return self.stages[name]

    def to_dict(self) -> dict[str, Any]:
        """Return dict of all this dataclass fields."""
        return asdict(self)



import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from textwrap import dedent

import pytest
from ddeutil.io.files import EnvFl


@pytest.fixture(scope="module")
def env_path(test_path) -> Iterator[Path]:
    this_path: Path = test_path / "env"
    this_path.mkdir(parents=True, exist_ok=True)
    with open(this_path / ".env", mode="w", encoding="utf-8") as f:
        f.write(
            dedent(
                """TEST=This is common value test
            # Comment this line ...
            COMMENT_TEST='This is common value test'  # This is inline comment
            QUOTE='single quote'
            DOUBLE_QUOTE="double quote"
            PASSING=${DOUBLE_QUOTE}
            UN_PASSING='${DOUBLE_QUOTE}'
            """
            ).strip()
        )

    yield this_path / ".env"

    shutil.rmtree(this_path)


def test_files_open_env_read(env_path):
    assert {
        "TEST": "This is common value test",
        "COMMENT_TEST": "This is common value test",
        "QUOTE": "single quote",
        "DOUBLE_QUOTE": "double quote",
        "PASSING": "double quote",
        "UN_PASSING": "${DOUBLE_QUOTE}",
    } == EnvFl(path=env_path).read(update=False)


def test_files_open_env_read_update(env_path):
    EnvFl(path=env_path).read(update=True)

    assert os.getenv("TEST") == "This is common value test"
    assert os.getenv("COMMENT_TEST") == "This is common value test"
    assert os.getenv("DOUBLE_QUOTE") == "double quote"
    assert os.getenv("PASSING") == "double quote"
    assert os.getenv("UN_PASSING") == "${DOUBLE_QUOTE}"


def test_files_open_env_write(env_path):
    with pytest.raises(NotImplementedError):
        EnvFl(path=env_path).write({"TEST": "This is common value test"})
